/*
 * io_avr.inc
 *
 * Hilfsdefinitionen fuer die Kompatibilitaet zwischen STM32- und AVR-Code
 *
 * GPIO-Operationen fuer AVR
 *
 * Created: 22.02.2014
 *  Author: Nicolas
 */


#ifndef AVRCOM_IO_AVR_H
    #define AVRCOM_IO_AVR_H
	
    #ifndef AVR
		#error "This include file is for AVR only";
	#endif
	


    #include <avr/io.h>
    #include "../src_avrcom/platform.h"
    #include "../src_base/datatypes.h"
    #include "../src_base/assert.h"
    #include "../src_base/datatypes.h"
    



    /* Einzelnes oder mehrere Bits eines Ports setzen
     *
     * GPIOx
     * Pin: Bitmaske */
    static_always_inline void io_setBit(GPIO_t * const pPORTx, const IOPin_t Pin) 
    {
        *pPORTx |= Pin;
    }



    /* Einzelnes oder mehrere Bits eines Ports loeschen
     *
     * GPIOx
     * Pin: Bitmaske */
    static_always_inline void io_clearBit(GPIO_t * const pPORTx, const IOPin_t Pin) 
    {
        *pPORTx &= ~((uint8_t) Pin);
    }



    /* Einzelnes oder mehrere Bits eines Ports umschalten
     *
     * GPIOx
     * Pin: Bitmaske */
    static_always_inline void io_toggleBit(GPIO_t * const pPORTx, const IOPin_t Pin) 
    {
        #if (defined(__AVR_ATmega644__))
            // Pin in einem Takt toggeln fuer neuere AVRs 
		    if(pPORTx == &PORTA) 
            {
    		    PINA |= Pin;
		    }
		    else if (pPORTx == &PORTB) 
            {
    		    PINB |= Pin;
		    }
		    else if (pPORTx == &PORTC) 
            {
    		    PINC |= Pin;
		    }
		    else if (pPORTx==&PORTD) 
            {
    		    PIND |= Pin;
		    }

        #elif defined(__AVR_ATmega32__)
            // Generische Variante fuer alle AVR-MCUs
            *pPORTx ^= Pin;
        
        #else   
            // Fuer nicht eingetragene MCUs: Generische Variante
            #warning "Pin toggle function may be slower than necessary"
        
            *pPORTx ^= Pin;
        
        #endif
    }



    /* Alle acht zusammenhaengende Bits eines Ports setzen oder loeschen
     *
     * GPIOx
     * byte  : Zielwerte
     * offset: wird beim AVR ignoriert */
    static_always_inline void io_setByte(GPIO_t * const pPORTx, const uint8_t byte, const IOPin_t offset) 
    {
		(void) offset;
		*pPORTx = byte;
    }



    /* Alle acht zusammenhaengende Bits eines Ports lesen
     *
     * GPIOx
     * offset: wird ignoriert
     * return: gelesenes Byte */
    static_always_inline uint8_t io_getByte(GPIO_t * const pPORTx, const uint8_t offset) 
    {
		(void) offset;
        
        #if defined(__AVR_ATmega32__)
            // Lesen per Offset der Port-Adresse (nur ueberprueft fuer einzelne
            // MCU-Typen)
            return *(pPORTx + (&PINA - &PORTA));

        #else
            // Aussuchen der Adresse aus den Defines

            #warning "Port operation may be slow"

		    if(pPORTx == &PORTA) 
            {
			    return PINA; 
		    }
		    else if (pPORTx == &PORTB) 
            {
			    return PINB;
		    }
		    else if (pPORTx == &PORTC) 
            {
			    return PINC;
		    }
		    else if (pPORTx == &PORTD) 
            {
			    return PIND;
		    }
	        return 0;
        #endif
    }



    /* Logischen Wert eines Bits lesen
     * Liefert logische 1, wenn Bit gesetzt, aber nicht notwendigerweise 
     * 0x01. 
     *
     * GPIOx
     * bit: Bitmaske */
	static_always_inline uint8_t io_bitIsSet(GPIO_t * const pPORTx, const IOPin_t Pin)
	{
		if(pPORTx == &PORTA) 
        {
			return PINA & Pin;
		}
		else if (pPORTx == &PORTB) 
        {
			return PINB & Pin;
		}
		else if (pPORTx == &PORTC) 
        {
			return PINC & Pin;
		}
		else if (pPORTx == &PORTD) 
        {
			return PIND & Pin;
		}
		return 0;
	}		



    /* Einzelne Pins eines Ports per Bitmaske als frei floatenden Eingang
     * setzen. 
     *         
     * GPIOx
     * Pin: Bitmaske */
    static_always_inline void io_setInput(GPIO_t * const pPORTx, const IOPin_t Pin)
	{
		if(pPORTx == &PORTA) 
        {
			DDRA &= ~Pin;
		}
		else if (pPORTx == &PORTB) 
        {
			DDRB &= ~Pin;
		}
		else if (pPORTx == &PORTC) 
        {
			DDRC &= ~Pin;
		}
		else if (pPORTx == &PORTD) 
        {
			DDRD &= ~Pin;
		}
	}



    /* Einzelne Pins eines Ports per Bitmaske als Push-Pull-Ausgang setzen.
     * Die Ports-Bits werden dabei nicht beeinflusst, muessen also bei 
     * Bedarf vorher oder nachher einzeln gesetzt oder geloescht werden.
     *         
     * GPIOx
     * Pin: Bitmaske */
    static_always_inline void io_setOutput(GPIO_t * const pPORTx,const IOPin_t Pin)
        {
        #if defined(__AVR_ATmega32__)
            // Setzen per Offset der Port-Adresse (nur ueberprueft fuer einzelne
            // MCU-Typen)
            *(pPORTx + (&DDRA - &PORTA)) |= Pin;

            static_assert(&DDRB == &PORTB + (&DDRA - &PORTA),"GPIO address mismatch" );
            static_assert(&DDRC == &PORTC + (&DDRA - &PORTA),"GPIO address mismatch" );
            static_assert(&DDRD == &PORTD + (&DDRA - &PORTA),"GPIO address mismatch" );

        #else
            // Aussuchen der Adresse aus den Defines

            #warning "Port operation may be slow"

	        if(pPORTx == &PORTA) 
            {
		        DDRA  |= Pin;
	        }
	        else if (pPORTx == &PORTB) 
            {
		        DDRB  |= Pin;
	        }
	        else if (pPORTx == &PORTC) 
            {
		        DDRC  |= Pin;
	        }
	        else if (pPORTx == &PORTD) 
            {
		        DDRD  |= Pin;
	        }
        #endif

        }



    /* Einzelne Pins eines Ports per Bitmaske als Eingang mit Pull-Up
     * setzen. 
     *         
     * GPIOx
     * Pin: Bitmaske */
    static_always_inline void io_setInputPU(GPIO_t * const pPORTx,const IOPin_t Pin)
    {
	    if(pPORTx == &PORTA) 
        {
		    DDRA  &= ~Pin;
		    PORTA |= Pin;
	    }
	    else if (pPORTx == &PORTB) 
        {
		    DDRB  &= ~Pin;
		    PORTB |= Pin;
	    }
	    else if (pPORTx == &PORTC) 
        {
		    DDRC  &= ~Pin;
		    PORTC |= Pin;
	    }
	    else if (pPORTx == &PORTD) 
        {
		    DDRD  &= ~Pin;
		    PORTD |= Pin;
	    }
    }



#endif 
