/*
 * intmath.c
 *
 * Einfache Funktionen fuer die Integer-Mathematik: Saettigende Addition
 *
 * Created: 08.03.2014
 *  Author: Nicolas
 */ 


#include "assert.h"
#include "intmath.h"



/* Saettingende Addition
 * val: Summand 1
 * incr: Summand 2
 * return: Summe/UINT8_MAX/UINT8_MIN  */
static inline uint8_t satadd_uint8(uint8_t val, int8_t incr)
{
    int_fast16_t temp = (int_fast16_t) val + incr;

    if (temp > (int_fast16_t) UINT8_MAX) 
    {
        temp = UINT8_MAX;
    }
    else if (temp < 0) 
    {
        temp = 0;
    }
    return (uint8_t) temp;
}


/* Saettingende Addition
 * val: Summand 1
 * incr: Summand 2
 * return: Summe/INT8_MAX/INT8_MIN  */
static inline int8_t satadd_int8(int8_t val, int8_t incr)
{
    int_fast16_t temp = (int_fast16_t) val + incr;

    if (temp > INT8_MAX) 
    {
        temp = INT8_MAX;
    }
    else if (temp < INT8_MIN) 
    {
        temp = INT8_MIN;
    }

    return temp;
}


/* Saettingende Addition
 * val: Summand 1
 * incr: Summand 2
 * return: Summe/UINT16_MAX/UINT16_MIN  */
static inline uint16_t satadd_uint16(uint16_t val, int16_t incr)
{
    int_fast24_t temp = (int_fast24_t) val + incr;

    if (temp > UINT16_MAX) 
    {
        temp = UINT16_MAX;
    }
    else if (temp < 0) 
    {
        temp = 0;
    }

    return temp;
}


/* Saettingende Addition
 * val: Summand 1
 * incr: Summand 2
 * return: Summe/INT16_MAX/INT16_MIN  */
static inline int16_t satadd_int16(int16_t val, int16_t incr)
{
    int_fast24_t temp = (int_fast24_t) val + incr;

    if (temp > INT16_MAX) 
    {
        temp = INT16_MAX;
    }
    else if (temp < INT16_MIN) 
    {
        temp = INT16_MIN;
    }

    return temp;
}



/* Saettingende Multiplikation
 * val: Faktor 1
 * fac: Faktor 2
 * return: Produkt/INT16_MAX/INT16_MIN  */
int8_t satmul_int8(int8_t val, int8_t fac)
{
    int_fast16_t temp = (int_fast16_t) val * fac;

    if (temp > INT8_MAX) 
    {
        temp = INT8_MAX;
    }
    else if (temp < INT8_MIN) 
    {
        temp = INT8_MIN;
    }

    return temp;
}



/* Saettingende Multiplikation
 * val: Faktor 1
 * fac: Faktor 2
 * return: Produkt/INT16_MAX/INT16_MIN  */
int16_t satmul_int16(int16_t val, int16_t fac)
{
    int32_t temp = (int32_t) val * fac;

    if (temp > INT16_MAX) 
    {
        temp = INT16_MAX;
    }
    else if (temp < INT16_MIN) 
    {
        temp = INT16_MIN;
    }

    return temp;
}



/* Addition mit Begrenzung auf waehlbaren Bereich
 * val:  Summand 1
 * incr: Summand 2
 * min:  Untere Wertegrenze
 * max:  Obere Wertegrenze */
uint8_t limadd_uint8(uint8_t val, int8_t incr, uint8_t min, uint8_t max)
{
    assert(min<=max);

    // Sicherstellen, dass Wertebereich nicht verlassen wird
    val = satadd_uint8(val,incr);
    
    // Begrenzen auf Wunschwert
    if (val > max) 
    {
        val = max;
    }
    else if (val < min) 
    {
        val = min;
    }
    
    return val;
}


/* Addition mit Begrenzung auf waehlbaren Bereich
 * val:  Summand 1
 * incr: Summand 2
 * min:  Untere Wertegrenze
 * max:  Obere Wertegrenze */
int8_t limadd_int8(int8_t val, int8_t incr, int8_t min, int8_t max) 
{
    assert(min<=max);
    
    // Sicherstellen, dass Wertebereich nicht verlassen wird
    val = satadd_int8(val,incr);

    // Begrenzen auf Wunschwert
    if (val > max) 
    {
        val = max;
    }
    else if (val < min) 
    {
        val = min;
    }
    
    return val;
}


/* Addition mit Begrenzung auf waehlbaren Bereich
 * val:  Summand 1
 * incr: Summand 2
 * min:  Untere Wertegrenze
 * max:  Obere Wertegrenze */
uint16_t limadd_uint16(uint16_t val, int16_t incr, uint16_t min, uint16_t max)
{
    assert(min<=max);
    
    // Sicherstellen, dass Wertebereich nicht verlassen wird
    val = satadd_uint16(val,incr);

    // Begrenzen auf Wunschwert
    if (val > max) 
    {
        val = max;
    }
    else if (val < min) 
    {
        val = min;
    }
   
    return val;
}


/* Addition mit Begrenzung auf waehlbaren Bereich
 * val:  Summand 1
 * incr: Summand 2
 * min:  Untere Wertegrenze
 * max:  Obere Wertegrenze */
int16_t limadd_int16(int16_t val, int16_t incr, int16_t min, int16_t max)
{
    assert(min<=max);
    
    // Sicherstellen, dass Wertebereich nicht verlassen wird
    val = satadd_int16(val,incr);
 
     // Begrenzen auf Wunschwert
    if (val > max) 
    {
        val = max;
    }
    else if (val < min) 
    {
        val = min;
    }
    
    return val;
}


/* Saettingende Addition
 * val: Summand 1
 * incr: Summand 2
 * return: Summe/UINT32_MAX/UINT32_MIN  */
uint32_t satadd_uint32(uint32_t val, int32_t incr)
{
	if (incr<0) 
    {
		// Subtraktion
		uint32_t nincr = (uint32_t) (-incr);
		if( nincr > val ) 
        {
			// Saettigung
			return 0;
		}
		else 
        {
			// Differenz
			return val - nincr;
		}
	}
	else 
    {
		// Addition
		uint32_t pincr = (uint32_t) incr;
		if(UINT32_MAX-pincr < val) 
        {
			// Saettigung
			return UINT32_MAX;
		} 
        else 
        {
			// Summe
			return val + pincr;
		}
	}
}



/* Hilfsfunktion: Sinus-Tabelle in 4-Grad-Schritten:
 * 13*sin(phi) 
 *
 * phi:     Winkel in Grad
 * return:  round(13*sin(phi)) */
constflash int8_t ytbl13[24] = {0,1,2,3,4,4,5,6,7,8,8,9,10,10,11,11,12,12,12,13,13,13,13,13};
int8_t sin13(int16_t phi)
{
    // Auf Kreis beschraenken
    phi %= 360;

    // Negative Werte beruecksichtigen
    if (phi < 0) 
    {
        phi += 360;
    }
        
    if (phi <= 90) 
    {
        phi +=2;
        phi /=4;
        return pgm_read_byte(&ytbl13[phi]);
    }
    else if (phi <= 180) 
    {
        phi = 180-phi;
        phi +=2;
        phi /=4;
        assert(phi<=23);
        return pgm_read_byte(&ytbl13[phi]);
    }
    else if (phi <= 270) 
    {
        phi -= 180;
        phi +=2;
        phi /=4;
        return -pgm_read_byte(&ytbl13[phi]);
    }
    else 
    {
        phi = 360-phi;
        phi +=2;
        phi /=4;
        return -pgm_read_byte(&ytbl13[phi]);
    }
}

