/*
 * eeprom_24Cx.h
 *
 * Lesen und Schreiben eines externen EEPROMs 24C02 per I2C
 *
 * Created: 02.02.2014
 *  Author: DL1DOW
 */


#include <string.h>
#include <stdlib.h>
#include "../src_base/assert.h"
#include "../src_avrcom/delay.h"
#include "../src_i2c/i2c_generic.h"
#include "eeprom_24Cx.h"





/* Einzelnes Byte aus EEPROM auslesen
 * Nur getesten fuer EEPROMS vom Typ 24C02, bei groesseren muss die
 * Speicheradresse in die I2C-Adresse eingerechnet werden */
uint_fast8_t ee24Cx_read_byte(uint8_t addr, uint8_t *byte)
{
	uint_fast8_t status;

	if(addr>EEPROM_SIZE_BYTES) 
    {
		return ee_invalidAddress<<4;
	}

	// Kommando ohne Stopbit senden
	status = I2C_Sendbuffer(EEPROM_I2C_DEVICE,EEPROM_I2C_ADDRESS, &addr, 1, noStop);
	if (status != 0)
		goto error;

	status = I2C_Receivebuffer(EEPROM_I2C_DEVICE, EEPROM_I2C_ADDRESS, byte, 1);
	if (status != 0)
		goto error;

	return status; // = 0

	error:
	I2C_reset(EEPROM_I2C_DEVICE);
	return status;
}


/* Einzelnes Byte in EEPROM schreiben
 * Nur getesten fuer EEPROMS vom Typ 24C02, bei groesseren muss die
 * Speicheradresse in die I2C-Adresse eingerechnet werden */
uint_fast8_t ee24Cx_write_byte(uint8_t addr, uint8_t byte)
{
	uint_fast8_t status;
	uint8_t buf[2];
	buf[0] = addr & 0xFF;
	buf[1] = byte;

	if(addr>EEPROM_SIZE_BYTES) 
    {
		return ee_invalidAddress<<4;
	}

	// Kommando mit Stopbit senden
	status = I2C_Sendbuffer(EEPROM_I2C_DEVICE, EEPROM_I2C_ADDRESS, buf, 2, stop);
	if (status != 0)
		goto error;

	return status; // = 0

	error:
	I2C_reset(EEPROM_I2C_DEVICE); // STOP wuerde auch reichen
	return status;
}


/* Viele Bytes aus EEPROM auslesen
 * Nur getesten fuer EEPROMS vom Typ 24C02, bei groesseren muss die
 * Speicheradresse in die I2C-Adresse eingerechnet werden
 * size = 0 ist immer erfolgreich und fuehrt keine Busoperation aus */
uint_fast8_t ee24Cx_read_block(uint8_t *buffer, uint8_t addr, uint8_t size)
{
	uint_fast8_t status;

	if(size == 0) return ee_ok;

	// Auf gueltige Adresse pruefen
	if( (addr+size)>EEPROM_SIZE_BYTES ) 
    {
		return ee_invalidAddress<<4;
	}
    
	if( (addr)>EEPROM_SIZE_BYTES ) 
    {
		return ee_invalidAddress<<4;
	}

	// Beim STM32F10x schlaegt das erste Lesen mit mehr als einem Byte direkt
	// nach der Initialisierung fehl. Deswegen wird ein einzelnes Dummy-Byte
	// vorher gelesen und der Rueckgabestatus ignoriert:
    status = ee24Cx_read_byte(0, buffer);
    (void) status; 

	// Kommando ohne Stopbit senden
	status = I2C_Sendbuffer(EEPROM_I2C_DEVICE, EEPROM_I2C_ADDRESS, &addr, 1, noStop);
	if (status != 0)
		goto error;

	status = I2C_Receivebuffer(EEPROM_I2C_DEVICE, EEPROM_I2C_ADDRESS, buffer, size);
	if (status != 0)
		goto error;

	return status;

	error:
	I2C_reset(EEPROM_I2C_DEVICE); // STOP wuerde meist auch reichen
	return status;
}


/* Ganze Seite in EEPROM schreiben
 * Nur getestet fuer EEPROMS vom Typ 24C02, bei groesseren muss die
 * Speicheradresse in die I2C-Adresse eingerechnet werden. Nur einzelne
 * Speicherseiten (16 bytes) werden unterstuetzt.
 * Wegen des malloc duerfte es eine der langsamsten und speicherintensivsten
 * Funktionen des gesamten Projekts sein. */
static uint_fast8_t ee24Cx_write_page(uint8_t *buffer, uint8_t addr, uint8_t size)
{
	uint_fast8_t status, page_start, page_end;
	uint8_t * bigbuffer;

	// Block mit Groesse Null ist immer OK und wird nicht auf den Bus geschickt
	if(size==0) return ee_ok;

	// Pruefen, ob Block auf einer Seite liegt
	page_start = addr>>4;
	page_end   = (addr+size-1)>>4;
	if( !(page_start == page_end) ) 
    {
		return ee_invalidAddress<<4;
	}

	// Pruefen, ob Block gueltige Adresse ist
	if( addr+size>EEPROM_SIZE_BYTES ) 
    {
		return ee_invalidAddress<<4;
	}

	bigbuffer = malloc(size+1U);
	if(bigbuffer == NULL) 
    {
		// Ein fehlgeschlagenes malloc ist nichts, wovon
        // man sich sinnvoll erholen knnte.
		error("malloc failed");
		return EXIT_FAILURE; // Wird nie aufgerufen, ueberzeugt aber Code-Tools
		                     // davon, dass bigbuffer nie NULL ist.
	}
    
	*bigbuffer = addr;
	memcpy(bigbuffer+1, buffer, size);

	// Kommando mit Stopbit senden
	status = I2C_Sendbuffer(EEPROM_I2C_DEVICE, EEPROM_I2C_ADDRESS, bigbuffer,\
			size+1,stop);
	free(bigbuffer);
    
	if (status != 0)
		goto error;

	return status; // = 0

	error:
	I2C_reset(EEPROM_I2C_DEVICE);
	return status;
}


/* EEPROM blockweise beschreiben, indem passend auf die Speicherseiten verteilt
 * wird */
uint_fast8_t ee24Cx_write_block(uint8_t *buffer,uint8_t addr,uint8_t size)
{
	uint_fast8_t status, blocksize;


	// Block mit Groesse Null ist immer OK und wird nicht auf den Bus geschickt
	if(size == 0) return ee_ok;

	do 
    {
		blocksize = 16 - (addr & 0x0F); // Seitenende
		if( size<blocksize ) blocksize = size;

		status = ee24Cx_write_page(buffer, addr, blocksize);
		if(status) 
        {
			return status;
		}
        
		buffer += blocksize;
		addr += blocksize;
		size -= blocksize;

		// Wartezeit bei mehreren Seiten noetig.
		delay_ms(5);
	} while( blocksize );

	return status;
}


/* EEPROM blockweise beschreiben, wobei vorher ueberprueft wird, ob nicht
 * schon der passende Inhalt im EEPROM hinterlegt ist. */
uint_fast8_t ee24Cx_update_block(uint8_t *buffer, uint8_t addr, uint8_t size)
{
	uint8_t * readbuffer;
	uint_fast8_t status;
	int equals; // memcmp muss in 16-Bit-Variable speichern

	readbuffer = malloc(size);
	if(readbuffer == NULL) 
    {
		// Ein fehlgeschlagenes malloc ist nichts, wovon
        // man sich sinnvoll erholen knnte.
		error("malloc failed");
		return EXIT_FAILURE; // Wird nie aufgerufen, ueberzeugt aber Code-Tools
		                     // davon, dass readbuffer nie NULL ist.
	}
	
    status = ee24Cx_read_block(readbuffer, addr, size);
    if (status != ee_ok) 
    {
    	free(readbuffer);
    	return status;
    }
    else 
    {
    	equals = memcmp(buffer, readbuffer, size);
    	free(readbuffer);
    }

    if (equals == 0)
    {
    	return status = ee_ok;
    }        
    else 
    {
        return ee24Cx_write_block(buffer, addr, size);
	}
}
