/*
 * nvram.c
 *
 * Behandlung des nichtfluechtigen Speichers, insbesondere Lesen und Speichern
 * von Konfigurationsdaten. Es werden Pruefsummen gebildet und ueberprueft.
 * Das bedingt einen Overhead von 10 Bytes je Datensatz - bei der Bildung
 * von Adressen ist das zu beruecksichtigen.
 * 
 * Created: 04.05.2015
 *  Author: Nicolas
 */ 


#include <string.h>
#include <stdlib.h>
#include "../src_avrcom/platform.h"
#include "../src_base/assert.h"
#include "../src_base/datatypes.h"
#include "../src_misc/crc.h"
#include "nvram.h"

#ifdef PLATFORM_AVR
  #include <avr/eeprom.h>

#elif defined(PLATFORM_STM32)
  #include "../src_eeprom/eeprom_24Cx.h"

#elif defined(PLATFORM_PC)
  	#include "../src_eeprom/eeprom_24Cx.h"

#else
    #error "unknown Platform"
    
#endif  




/* | (uint32) 0xC0DEDBAD | .... Datensatz .... | CRC16 | */
#define NVRAM_OVERHEAD 10U






/* Datensatz aus dem EEPROM lesen und ueberpruefen. Der Puffer ist im
 * nicht-erfolgreichen Fall mit ungueltigen Daten ueberschrieben.
 *
 * *buffer: Puffer fuer Rueckgabe, min. Groesse "size"
 * address: Speicheradresse des gewuenschten Datensatzes
 * size   : Groesse des gewuenschten Datensatzes
 * return : EXIT_SUCCESS oder EXIT_FAILURE */
uint_fast8_t nvram_read(uint8_t *buffer, uint_fast8_t address, uint_fast8_t size)
{
	// "crc1" und "akindofmagic" werden als Puffer fester Breite verwendet.
	uint16_t crc1, crc2;
	uint32_t akindofmagic;

    #ifdef AVR
        // Auf dem AVR sind die EEPROM-Routinen zeigerbasiert
        uint8_t *addr = (uint8_t*)(size_t) address;
        
    #else
        // Ueberall sonst sind die EEPROM-Adressen normale Integer
        uint_fast8_t addr = address;
        uint_fast8_t status = 0;
        
    #endif

	// Daten aus EEPROM auslesen. Fehlercode wird nicht ausgewertet, da:
    //  - Die AVR-EEPROM-Routinen keinen Rueckgabewert haben und
    //  - Fehler  beim Auslesen aus dem Datensatz selbst erkannt werden.
	#ifdef AVR
		eeprom_read_block((uint8_t *) &akindofmagic,addr,4);
		eeprom_read_block(buffer,addr+4,size);
		eeprom_read_block((uint8_t *) &crc1,addr+4+size,2);
        
	#else
		status |= eeprom_read_block((uint8_t *) &akindofmagic,addr,4);
		status |= eeprom_read_block(buffer,addr+4,size);
		status |= eeprom_read_block((uint8_t *) &crc1,addr+4+size,2);

		if(status) 
        {
			return nvram_i2c_failure;
		}
        
	#endif

	// Die ersten 32 Bit des Dataset muessen ein magisches Wort sein
	if (akindofmagic != MAGICNUMBER) 
    {
		return nvram_magic_failure;
	}

	// Ueber den restlichen Teil CRC berechnen. Die letzten zwei Bytes sind die
	// CRC des Datensatzes, also muss der Gesamt-CRC Null ergeben.
	crc2 = crc16_calc(buffer, size);
	if(crc1 != crc2) 
    {
		return nvram_crc_failure;
	}
	else 
    {
		return nvram_ok;
	}
}


/* Konfigurationsdaten in EEPROM schreiben. Vorher wird eine Pruefsumme
 * gebildet und nach dem Schreiben ueberprueft.
 * Wenn das schreiben fehlerhaft war, sind die alten Konfigurationsdaten im
 * EEPROM leider verloren.
 *
 * *dataset: Zu schreibender Datensatz
 * address : gewuenschte Speicheradresse
 * size    : Groesse des Datensatzes in Bytes */
uint_fast8_t nvram_write(uint8_t *dataset, uint_fast8_t address, uint_fast8_t size)
{
	// "magic", "crc" und "buffer" werden als Puffer fester Breite verwendet.
	uint32_t magic = MAGICNUMBER;
	uint16_t crc;
    uint8_t *buffer;
    uint_fast8_t blksize = 0, status;
    
    #ifdef AVR
        // Auf dem AVR sind die EEPROM-Routinen zeigerbasiert
        uint8_t *addr = (uint8_t*)(size_t) address;
        
    #else
        // Ueberall sonst sind die EEPROM-Adressen normale Integer
        uint_fast8_t addr = address;
        
    #endif

    // Die ersten 32 Bit des Dataset sind ein magisches Wort
	blksize = 4;
    eeprom_update_block((uint8_t *) &magic, addr, blksize);
 	addr += blksize;

	// Es folgt der eigentliche Datensatz
	blksize = size;
	eeprom_update_block(dataset, addr, blksize);
	addr += blksize;

	// Ueber den Datensatz (ohne Magisches Wort) wird eine CRC16 berechnet
	// und angehaengt
	blksize = 2;
    crc = crc16_calc(dataset, size);
	eeprom_update_block((uint8_t *) &crc, addr, blksize);
	addr += blksize;

	// Der Erfolg der Schreiboperation wird ueberprueft, indem der Datensatz
	// noch einmal eingelesen wird.
	buffer = malloc(size);
	if( buffer == NULL ) 
    {
		// Ein fehlgeschlagenes malloc ist nichts, wovon
        // man sich sinnvoll erholen knnte.
		error("malloc failed");
		return EXIT_FAILURE; // Wird nie aufgerufen, ueberzeugt aber Code-Tools
		                     // davon, dass buffer nie NULL ist.
	}

	status =  nvram_read(buffer, address, size);
	if ( status != EXIT_SUCCESS ) 
    {
		return EXIT_FAILURE;
	}

	// Der Check in edm_nvram_read sollte jeden Lesefehler ausschliessen -
	// schlaegt der Vergleich mit dem urspruenglichen Datensatz also fehl,
	// liegt ein Programmfehler vor.
	status = (memcmp(buffer, dataset,size) != 0);
	free(buffer);
	if ( status ) 
    {
		error("Error cheking NRAM data");
	}

    return EXIT_SUCCESS;
}




