/*
 * encoder_panasonic.c
 *
 * Auslesen eines Drehgebers
 *
 * Die Funktion encoder_get() pollt den Drehgeber ohne !negative!
 * Nebenwirkungen und kann beliebig oft aufgerufen werden. encoder_get_diff()
 * gibt Werte zwischen aufeinanderfolgenden Aufrufen zurueck.
 * Mit encoder_get() in einer schnellen Programmschleife oder ISR und
 * encoder_get_diff() in einer langsamen Programmschleife laesst sich eine
 * schnelle Benutzerschnittstelle realisieren.
 *
 * Funktionen fuer AVR und STM32
 *
 * Created: 19.12.2013
 *  Author: Nicolas
 */ 



#include "encoder_panasonic.h"
#include "../src_avrcom/platform.h"

#ifdef AVR
    /* Die Pin-Namen sind beim AVR etwas anders als beim STM32. Alle anderen 
    AVR-spezifischen Defines sind schon in gpio.h und pgmspace.h */

	#define ENCODER_SOFT
    
#elif defined STM32F10x
    #include "stm32f10x_gpio.h"
    #include "stm32f10x_rcc.h"
    #include "stm32f10x_tim.h"

	#ifndef ENCODER_HARD
		#define ENCODER_SOFT
	#endif
    
#elif defined STM32F4XX
    #define STM32F4xx
    #include "stm32f4xx_gpio.h"

#elif defined(PLATFORM_PC)
	#include <stdbool.h>
	#include <stdint.h>
	#include "../src_mockup_sdl/mock_input.h"
	#include "../src_mockup_sdl/mock_sdl.h"

	#define ENCODER_MOCK

#else
    #error "unknown MCU"
    
#endif



#include "../src_base/assert.h"
#include "../src_avrcom/io.h"
#include "../src_avrcom/pgmspace.h"
#include "../../ioMapping.h"

   






#ifdef ENCODER_SOFT
    /* 
     * Auslesen eines Drehgebers mit reinen GPIO-Operationen
     * 
     * Unterstuetzt werden auch Panasonic-Encoder mit wackligen Rastpunkten auf
     * Spur B.
     * Implementierung nach Vorschlag von Artikel "Drehgeber" auf
     * https://www.mikrocontroller.net/articles/Drehgeber
     */


    // Dekodertabelle fr wackeligen Rastpunkt (halbe Aufloesung)
    constflash int_fast8_t table[16] = {0,0,-1,0,0,0,0,1,1,0,0,0,0,-1,0,0};

    // Dekodertabelle fr normale Drehgeber (volle Aufloesung)
    //constflash int_least8_t table[16] = {0,1,-1,0,-1,0,0,1,1,0,0,-1,0,-1,1,0};


    /* I/O einstellen */
    void encoder_init(void)
    {
        io_setInputPU(ENC_A_GPIO, ENC_A_Pin);
        io_setInputPU(ENC_B_GPIO, ENC_B_Pin);
    }


    /* Drehgeber pollen, Absolutwert abfragen und Statusupdate vollziehen. Die
     * Differenz zum vorherigen Aufruf ist nie groesser als 1. Haeufige Aufrufe
     * haben keine negative Nebenwirkungen, deshalb eignet sie sich fuer
     * Status-Updates in Interrupt-Routinen und Programmschleifen gleichermassen.
     *
     * Fuer die Panasonic-Encoder als Eingabegeraet sollte der Drehgeber 
     * mindestens alle 2 Millisekunden gepollt werden.
     *
     * return: Absolutwert */
    uint_fast8_t encoder_get(void)
    {
	    static uint_fast8_t enc_savebits = 0x01;
	    static uint_fast8_t enc_val = 0x00;

	    enc_savebits = (enc_savebits << 2)  & 0x0F;
	    if( io_bitIsSet(ENC_A_GPIO, ENC_A_Pin) ){ enc_savebits |=2; }
	    if( io_bitIsSet(ENC_B_GPIO, ENC_B_Pin) ){ enc_savebits |=1; }
	    enc_val -= (uint_fast8_t) table[enc_savebits];

	    return enc_val;
    }
    
 

#elif defined STM32F10x && defined ENCODER_HARD   
    /*
     * Nutzung des Hardware-Encoders im STM32F10x
     * Fuer andere MCUs nicht implementiert */

    // Pins und Timer einstellen
    void encoder_init(void)
    {
        GPIO_InitTypeDef GPIO_InitStructure;

        RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOA | RCC_APB2Periph_AFIO,ENABLE);
        RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM3,ENABLE);

        GPIO_InitStructure.GPIO_Pin   = GPIO_Pin_6 | GPIO_Pin_7;
        GPIO_InitStructure.GPIO_Mode  = GPIO_Mode_IPU; // In, pull-up
        GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;

        GPIO_Init(ENC_A_GPIO,&GPIO_InitStructure);

        TIM_EncoderInterfaceConfig(TIM3, TIM_EncoderMode_TI12, TIM_ICPolarity_Rising, \
            TIM_ICPolarity_Falling);
        TIM_Cmd(TIM3, ENABLE);
    }


    /* Hardware-Encoder aus dem Device auslesen */
    uint_fast8_t encoder_get(void) 
    {
	    return (uint_fast8_t) TIM_GetCounter(TIM3);
    }



#elif defined STM32F4xx && defined ENCODER_HARD
    /* 
     * Hardware-Unterstuetzung momentan noch nicht implementiert 
     */
    #error "Hardware Encoder support not implemented for this device yet";


#elif defined ENCODER_MOCK
    /* 
     * Bereitstellung der Encoder-Funktionen fuer einen Mockup mit SDL/Matlab
     * Der Werte "sdlenc" muss extern zur Verfuegung gestellt werden, hier wird
     * er nur in die passende Funktion gekleidet.
     */
    
    /* Pins einstellen (Dummy), hier: Wert ruecksetzen */
    void encoder_init(void)
    {
        sdlenc = 0;
    }


    uint_fast8_t encoder_get(void)
    {
	    /* Encoder-Wert updaten
	     * sowohl pollen fuer internen Statusupdate als auch Auslesen der internen
	     * Werte */
        mock_event_poll();
	    return sdlenc;
    }

#endif



/* Differenz des Drehgeber-Standes zwischen aufeinanderfolgenden Aufrufen der
 * Funktion auslesen.
 *
 * return: Differenz zu vorherigem Aufruf */
int_fast8_t encoder_get_diff(void)
{
	static uint_fast8_t enc_last;
	uint_fast8_t enc_this;
	int_fast8_t enc_diff;

	static_assert(sizeof(uint_fast8_t)==sizeof(int_fast8_t), \
			"variable must be of same size");

	enc_this = encoder_get();
	enc_diff = (int_fast8_t) (enc_this-enc_last);
	enc_last = enc_this;

	return enc_diff;
}

