﻿/*
 * glcd_gfxfun.h
 *
 * Grafik-Funktionen (Kreise, Linien, Bitmaps etc. fuer Grafik-LCD)
 *
 * Created: 15.02.2016 08:36:50
 *  Author: Nicolas
 */ 



#include "./glcd_gfxfun.h"
#include <stdlib.h>
#include "./glcd_generic.h"
#include "../src_base/assert.h"
#include "../src_base/intmath.h"



/* Bitmap (z.B. Icon) aus dem Flash an der aktuellen Position darstellen
 * Die Bitmap muss so aufgebaut sein, dass an Stelle 0 die Breite und an Stelle
 * 1 die Hoehe steht. */
void glcd_putbitmap_P(constflash unsigned char * bitmap)
{
    uint16_t pos0 = glcd_getpos();
    uint8_t yPx  = bitmap[0];   // Erstes Zeichen Breite (Pixel)
    uint8_t xBlk = bitmap[1]/8; // Zweites Zeichen Hoehe (Pixel)
    uint16_t idx = 2;
    uint8_t i;
    
    if(bitmap==NULL) return;
    
    for ( ; xBlk>0; xBlk--) 
    {// X geht nach unten
        for (i = yPx; i>0; i--) 
        {
            glcd_putdat(bitmap[idx++]);
        }
        glcd_setpos(pos0 += 128);
    }
}



/* Linie zeichnen nach dem Bresenham-Algorithmus
 * Implementierung nach dem gleichnamigen Wikipedia-Artikel 
 *
 * x0, y0: Startpunkt
 * x1, y1: Zielpunkt */
void glcd_line(uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1)
{
    int_fast8_t dx =  ABS(x1-x0), sx = x0<x1 ? 1 : -1;
    int_fast8_t dy = -ABS(y1-y0), sy = y0<y1 ? 1 : -1;
    int_fast8_t err = dx+dy, e2; /* error value e_xy */

    while(1)
    {
        glcd_setPixel(x0, y0);
        if (x0 == x1 && y0 == y1) break;
        e2 = 2*err;
        if (e2 > dy) { err += dy; x0 += sx; } /* e_xy+e_x > 0 */
        if (e2 < dx) { err += dx; y0 += sy; } /* e_xy+e_y < 0 */
    }
}




/* ***********************************************************************************************
 *       Kreise / Kreisabschnitte plotten                                                        *
 * ********************************************************************************************* */

/* Hilfsfunktion: Vollstaendige Kreissegmente (Achtelkreise) nach dem Bresenham-Algorithmus plotten
 * Implementierung nach Wikipedia 
 * x0, y0: Kreismittelpunkt, radius: Radius
 * seg: bitfeld mit 1: Achtelkreisabschnitt wird geplottet, 0: Abschnitt fehlt. */
static void segment_full(int x0, int y0, int radius, uint8_t seg)
{
    int f = 1 - radius;
    int ddF_x = 0;
    int ddF_y = -2 * radius;
    int x = 0;
    int y = radius;
    
    
    while( x < y )
    {
        if( f >= 0 )
        {
            y--;
            ddF_y += 2;
            f += ddF_y;
        }
        x++;
        ddF_x += 2;
        f += ddF_x + 1;

        if ( seg & 1U<<0 ) glcd_setPixel(x0 - y, y0 + x); // 0. Achtelkreisabschnitt
        if ( seg & 1U<<1 ) glcd_setPixel(x0 - x, y0 + y); // 1. Achtelkreisabschnitt
        if ( seg & 1U<<2 ) glcd_setPixel(x0 + x, y0 + y); // 2. Achtelkreisabschnitt
        if ( seg & 1U<<3 ) glcd_setPixel(x0 + y, y0 + x); // 3. Achtelkreisabschnitt
        if ( seg & 1U<<4 ) glcd_setPixel(x0 + y, y0 - x); // 4. Achtelkreisabschnitt
        if ( seg & 1U<<5 ) glcd_setPixel(x0 + x, y0 - y); // 5. Achtelkreisabschnitt
        if ( seg & 1U<<6 ) glcd_setPixel(x0 - x, y0 - y); // 6. Achtelkreisabschnitt
        if ( seg & 1U<<7 ) glcd_setPixel(x0 - y, y0 - x); // 7. Achtelkreisabschnitt
    }
}



/* Hilfsfunktion: Einen Unvollstaendigen Achtelkreisabschnitt plotten 
 * ebenfalls in Anlehnung an den Wikipedia-Artikel zum Bresenham-Algorithmus */
static void segment_part(int x0, int y0, int radius, int ang0, int ang1, uint8_t oct)
{
    assert(radius==13);
    
    int f = 1 - radius;
    int ddF_x = 0;
    int ddF_y = -2 * radius;
    int x = 0;
    int y = radius;
    uint8_t cos0, cos1;
    
    oct%=8;

    if (oct%2) 
    {
        // Rueckwartszaehlende Quadranten
        cos0 = sin13(45) - sin13(ang1);
        cos1 = sin13(45) - sin13(ang0);
    }
    else 
    {
        // Vorwaertszaehlende Quadranten
        //  sin(0° < phi < 45°) > 0
        cos0 = abs(sin13(ang0));
        cos1 = abs(sin13(ang1));
    }
    
    while(x < y)
    {
        if(f >= 0)
        {
            y--;
            ddF_y += 2;
            f += ddF_y;
        }
        x++;
        ddF_x += 2;
        f += ddF_x + 1;
        

        if ( (x > cos0) && (x <= cos1) ) 
        {
            switch (oct) 
            {
                case 0:
                    glcd_setPixel(x0 - y, y0 + x); // 0. Achtelkreissegment
                    break;
                case 1:
                    glcd_setPixel(x0 - x, y0 + y); // 1. Achtelkreissegment
                    break;
                case 2:
                    glcd_setPixel(x0 + x, y0 + y); // 2. Achtelkreissegment
                    break;
                case 3:
                    glcd_setPixel(x0 + y, y0 + x); // 3. Achtelkreissegment
                    break;
                case 4:
                    glcd_setPixel(x0 + y, y0 - x); // 4. Achtelkreissegment
                    break;
                case 5:
                    glcd_setPixel(x0 + x, y0 - y); // 5. Achtelkreissegment
                    break;
                case 6:
                    glcd_setPixel(x0 - x, y0 - y); // 6. Achtelkreissegment
                    break;
                case 7:
                    glcd_setPixel(x0 - y, y0 - x); // 7. Achtelkreissegment
            }
        }        
    }
}



/* Kreis oder Kreisabschnitt plotten
 * Implementierung des Bresenham-Algorithmus frei nach Wikipedia
 * x0, y0: Kreismittelpunkt, radius: Radius
 * ang0, ang1: Start und Endwinkel (Reihenfolge ist egal, der kleinere ist immer Start)
 *             Ist die Differenz zwischen ang0, ang1 > 360°, wird ein Vollkreis dargestellt.
 */
void glcd_circle(int x0, int y0, int radius, int ang0, int ang1)
{
    int temp;
    uint8_t voct = 0x00; // Vollstaendige Achtelkreissegmente
    uint8_t vedg = 0x00; // Vollstaendige "Ecken"
    int8_t oct0, oct1; // Anfangs- und End-Achtelkreis
    int8_t rem0, rem1; // Restkreisabschnitt, der kein vollstaendiger Achtelkreis ist


    // Reihenfolge sicherstellen
    if ( ang1 < ang0 ) 
    {
        temp = ang1;
        ang1 = ang0;
        ang0 = temp;
    }

    // Spezialfall: geschlossener Kreis
    if ( ang0 + 360 <= ang1 ) 
    {
        ang0 = 0;
        ang1 = 360;
    }

    // Winkel begrenzen (-360 -- 360°)
    ang0 = ang0 >  360 ?  360 : ang0;
    ang0 = ang0 < -360 ? -360 : ang0;
    ang1 = ang1 >  360 ?  360 : ang1;
    ang1 = ang1 < -360 ? -360 : ang1;

    
    // Feststellen, welche Oktanten und Ecken gefuellt sind
    
    // oct0,oct1 : -8...8 -> 0...16
    oct0 = ((ang0+360)/45);
    rem0 = (ang0+360)%45;
    oct1 = ((ang1+360)/45);
    rem1 = (ang1+360)%45;
    
    
    // Vollstaendige Segmente und Ecken identifizieren
    for ( int8_t i = 0; i <= 16; i++ ) 
    {
        if ( (oct0 <= i-1 && oct1 >= i+1) ) voct |= 1U<<(i)%8; // 8 Segmente
        if ( (oct0 <= i && oct1 >= i+1)   ) vedg |= 1U<<(i+1)%8; // 8 "Ecken"
    }

    
    // Luecken fuellen (Punkte an NO, SO, SW, NW muessen nicht gefuellt werden)
    if ( vedg & 1U<<0 ) glcd_setPixel(x0 - radius, y0); // N
    if ( vedg & 1U<<2 ) glcd_setPixel(x0, y0 + radius); // O
    if ( vedg & 1U<<4 ) glcd_setPixel(x0 + radius, y0); // S
    if ( vedg & 1U<<6 ) glcd_setPixel(x0, y0 - radius); // W
    

    // Vollstaendige Segmente zeichnen
    segment_full(x0, y0, radius,voct);
    
    
    // Unvollstaendige Segmente zeichnen
    if (oct0 == oct1) 
    {
        segment_part(x0, y0, radius, rem0, rem1, oct0 & ~INT8_MIN);
    }
    else 
    {
        // oct0, oct1 > 0, & ~INT8_MIN unterdrueckt die Warnung
        segment_part(x0, y0,radius, 0, rem1, oct1 & ~INT8_MIN);
        segment_part(x0, y0,radius, rem0, 45, oct0 & ~INT8_MIN);
    }
}
