/*
 * Ansteuerung Grafik-LCD TG12864B
 * memory-mapped
 * SRAM-Verbrauch ueber 1kb!
 *
 * Created: 25.08.2013
 *  Author: Nicolas
 */ 

#include <stdint.h>
#include "glcd_ks0108.h"
#include "../src_base/datatypes.h"


/* Displaymasse
 * Eizelne Bloecke werden wie Zeilen behandelt */

#define GLCD_KS0108_CHARWIDTH 5U
#define GLCD_KS0108_DISPWIDTH 64U
#define GLCD_KS0108_DISPHEIGHT 16U
#define GLCD_KS0108_DRAW_DELAY 50U  // Anzahl der Mindestaufrufe von
                                    // glcd_ks0108_timer(), bevor wieder
                                    // neugezeichnet wird.
#define LEFT  0x00
#define RIGHT 0x01


static uint16_t gl_pos = 0;
#ifdef GLCD_KS0108_FRAMEBUFFER
    uint8_t gl_framebuffer[1024];
    
    // Startwert unerlaubter Wert, wenn er gelesen wird, ist die Bremse nie aktiv gewesen:
    static uint8_t gl_delay_count = GLCD_KS0108_DRAW_DELAY+1; 
#endif


/* ************************************************************************************ *
 *     Prozessorspezifische Funktionen                                                  *
 * ************************************************************************************ */

// Ports auf Ausgang
extern void glcd_ks0108_setIo(void);

/* Zeichen oder Kommando an Display senden *
 * RS : 0=Instruction 1=data               */
extern void glcd_ks0108_putsth(uint8_t value,uint8_t RS);

/* Displayseite waehlen (L,R) */
extern void glcd_ks0108_area(uint8_t LR);

// Sicherheitsabstand zu Power on
extern void glcd_ks0108_startdelay(void);

// GLCD reseten gemaess Datenblatt S. 16
extern void glcd_ks0108_reset(void);


/* ************************************************************************************ *
 *  Generische Displayfunktionen                                                        *
 * ************************************************************************************ */
/* Kommando an Display senden */
static_inline void glcd_ks0108_putcom(uint8_t value) 
{
    glcd_ks0108_putsth(value,0x00);
}



/* Byte an Display zur Darstellung senden */
#ifdef GLCD_KS0108_FRAMEBUFFER
    void glcd_ks0108_putdat(uint8_t value) 
	{
        gl_framebuffer[gl_pos] = value;
        gl_pos++;
        gl_pos%=GLCD_KS0108_END;
    }
#else

    void glcd_ks0108_putdat(uint8_t value) 
	{
	    glcd_ks0108_putsth(value,0x01);
	    // Positionszaehler inkrementieren; wenn Displayblock ueberschritten in naechste 
        // Position springen.
	    gl_pos++;
	    if(!(gl_pos%GLCD_KS0108_DISPWIDTH)) 
        {
		    glcd_ks0108_setpos(gl_pos);
	    }
    }
    
#endif 


/* Byte mit bestehendem Bildinhalt verschmelzen */
#ifdef GLCD_KS0108_FRAMEBUFFER
    void glcd_adddat(uint8_t value)
	{
        gl_framebuffer[gl_pos] |= value;
        gl_pos++;
        gl_pos%=GLCD_KS0108_END;
    }
    
#else
    void glcd_ks0108_adddat(uint8_t value) 
	{
		#warning "glcd_ks0108_adddat() has never been tested"
	    ret = glcd_ks0108_readdat(void);
		// keine Ahnung, ob readdat einen Vorschub macht
		glcd_ks0108_putdat(ret | value);
    }
    
#endif 



/* Einzelnes Pixel setzen
 * Die Funktion funktioniert bislang nur bei der Variante mit Framebuffer. 
 * x: vertikale Koordinate (nach unten)
 * y: horizontale Koordinate (nach rechts)
 * return: Urspruenglicher Pixelwert */
uint8_t glcd_ks0108_setPixel(uint8_t x, uint8_t y)
{
    uint8_t ret;
    #ifdef GLCD_FRAMEBUFFER
		uint_fast16_t byteaddr = x/8*128U+y;
		ret = gl_framebuffer[byteaddr];
		gl_framebuffer[byteaddr] |= 0x1 << (x%8);
	#else
		
		#warning "This function has never been tested without FRAMEBUFFER"
		// und langsam ist es wahrscheinlich auch.
		size_t byteaddr = x/8*128+y;
		glcd_ks0108_setpos(byteaddr);
		ret = glcd_ks0108_readdat(void);
		glcd_ks0108_putdat(ret | (0x1 << (x%8));
    #endif
    
    return ret;
}


/* Der Inhalt des Displays wird geloescht, indem das ganze Bild mit leeren Pixeln 
 * ueberschrieben wird. */
#ifdef GLCD_KS0108_FRAMEBUFFER
    void glcd_ks0108_clear(void) 
	{
        uint_fast16_t i;
        for (i = 0;i<GLCD_KS0108_END;i++) 
        {
            gl_framebuffer[i] = 0x00;
        }
        
        // Aus Kompatibilitaetsgruenden zu direktem Schreiben:
        glcd_ks0108_setpos(0); 
    }
    
#else
    void glcd_ks0108_clear(void) 
	{
        uint_fast16_t i;  
        for(i = 0; i<GLCD_KS0108_END; i++) 
        {
	        glcd_ks0108_putdat(0x00);
        }	  	
        glcd_ks0108_setpos(0);
    }
    
#endif





/* Initialisierung: Ports vorbereiten, Display einschalten und loeschen       */
void glcd_ks0108_init()
{
    glcd_ks0108_startdelay(); // 50ms, Sicherheitsabstand zu Power on
    glcd_ks0108_setIo();
    glcd_ks0108_reset();

    // Display einschalten (links und rechts)
    glcd_ks0108_area(LEFT);
    glcd_ks0108_putcom(0x3F); // Display ON:         0b00111111
    glcd_ks0108_putcom(0xC0); // Erste Zeile Anfang: 0b11000000
    glcd_ks0108_putcom(0xB8); // Erste Zeile (X):    0b10111000
    glcd_ks0108_putcom(0x40); // Erste Spalte (Y):   0b01000000

    glcd_ks0108_area(RIGHT);
    glcd_ks0108_putcom(0x3F); // Display ON
    glcd_ks0108_putcom(0xC0); // Erste Zeile
    glcd_ks0108_putcom(0xB8); // Erste Zeile (X)
    glcd_ks0108_putcom(0x40); // Erste Spalte (Y)

    glcd_ks0108_clear();
    glcd_ks0108_drawnow();
    gl_pos = 0;
}


uint16_t glcd_ks0108_getpos(void) 
{
    return gl_pos;
}


extern uint8_t glcd_ks0108_readdat();


/* Absolute position from 0 to 1024 */
#ifdef GLCD_KS0108_FRAMEBUFFER
    void glcd_ks0108_setpos(uint16_t pos)
    {
        gl_pos = pos % GLCD_KS0108_END;
    }
       
#else
    void glcd_ks0108_setpos(uint16_t pos) 
    {
        uint8_t PosY = pos%64;  // y geht nach rechts
	    uint8_t block = pos/64; // 0 .. 15, zwei pro Zeile

      
        pos%=GLCD_KS0108_END;
      
	    if(!(block%2)) 
        {
		    glcd_ks0108_area(LEFT);
	    } 
        else 
        {
		    glcd_ks0108_area(RIGHT);
		    block = block-1;
	    }
	    block /= 2;		

	    block%=8;                         // erlaubter Bereich
	    glcd_ks0108_putcom(0xB8|(block)); //X-Adresse-Reihe:  0b10111000|(PosX)
	    glcd_ks0108_putcom(0x40|PosY);    //Y-Adresse,Spalte: 0b01000000|(PosY)
	    gl_pos = pos;
    }
    
#endif

#ifdef GLCD_KS0108_FRAMEBUFFER
    /* Display-Pointer fuer direktes Schreiben in den Framebuffer */
    uint8_t * glcd_ks0108_getmemory(void) 
	{
        return gl_framebuffer;
    }
    
#endif


#ifdef GLCD_KS0108_FRAMEBUFFER
    // Framebuffer Gelegenheit zum Update geben; Update wird nur ausgefuehrt, wenn der
    // Verzoegerungszaehler auf Null ist
    void glcd_ks0108_draw(void) 
	{
    
        // Verzoegerung pruefen
        if (gl_delay_count == GLCD_KS0108_DRAW_DELAY+1)
        {
            // Verzoegerung inaktiv, immer neuzeichnen
        }            
        else if (gl_delay_count)
        {
            // Verzoegerung aktiv, jetzt nicht dran
            return;
        }            
        else 
        {
            // Verzoegerung aktiv, jetzt neuzeichnen
            gl_delay_count = GLCD_KS0108_DRAW_DELAY;
        }

        glcd_ks0108_drawnow();
    } 
#else
    // Funktion ist ueberfluessig bei direktem Schreiben
    void glcd_ks0108_draw(void) {}
        
#endif  


  /* Framebuffer direkt auf Display ausgeben */
void glcd_ks0108_drawnow(void) 
{
    glcd_ks0108_drawField(0U);
    glcd_ks0108_drawField(1U);
} 



/* Eine Haelfte des Framebuffers direkt auf Display ausgeben */
#ifdef GLCD_KS0108_FRAMEBUFFER
    void glcd_ks0108_drawField(uint_fast8_t fieldNo)
    {
        uint_fast16_t idx;
        uint8_t i, j, tmp;
      
        
        // Auf Nullposition (low level), framebuffer ausgeben
        if (fieldNo == LEFT)
        {
            glcd_ks0108_area(LEFT);    
            idx = 0U;
        }
        else
        {
            glcd_ks0108_area(RIGHT);
            idx = 64U;
        }
      
      
        // Nur fuer den Fall, dass bei der vorherigen Uebertragung nicht die korrekte 
        // Anzahl uebermittelt wurde
        glcd_ks0108_putcom(0x40|0); //Y-Adresse,Spalte: 0b01000000|(PosY)
      
      
        for (i = 0;i<GLCD_KS0108_DISPHEIGHT/2;i++) 
        {
            glcd_ks0108_putcom(0xB8|i); //X-Adresse-Reihe:  0b10111000|(PosX)
            for (j = 0; j<64; j++) 
            {
                tmp = gl_framebuffer[idx++];
                glcd_ks0108_putsth(tmp,0x01);
            }
            #pragma GCC diagnostic push
            #pragma GCC diagnostic ignored "-Wconversion"
                idx += GLCD_KS0108_DISPWIDTH;
            #pragma GCC diagnostic pop
        }
      

        // LCD wieder auf hinterlegte Position springen fuer normale Ausgabe
        glcd_ks0108_setpos(gl_pos);
    } // void glcd_ks0108_drawnow(void)
    
#else
    // Ohne Framebuffer ist die Funktion voellig nutzlos, da immer direkt
    // auf das Display geschrieben wird.
    void glcd_ks0108_drawField (uint8_t fieldNo) 
    {
        (void) fieldNo;
    }
    
#endif // ifdef GLCD_KS0108_FRAMEBUFFER


// Delay-Zaehler fuer Aufruf von Drawnow; zaehlt immer auf Null herunter und wartet 
// darauf, von draw-Routine wieder zurueckgesetzt zu werden
#ifdef GLCD_KS0108_FRAMEBUFFER
    void glcd_ks0108_timer(void) 
	{
        if (gl_delay_count) gl_delay_count--;
    }
    
#else
    void glcd_ks0108_timer(void) { }
        
#endif
