/*
 * glcd_textfun.c
 *
 * Textfunktionen fr Grafikdisplays mit nur einem Display-Controller, d.h. durchgehend 
 * adressiertem SRAM
 * Die Darstellung von Text funktioniert bislang nur im logischen Zeilenraster des LCDs 
 * (d.h. Schrifthhe max. 7 Pixel und nur im 8-Pixel-Raster positionierbar).

 * Allein eine 5x7-Schriftart einzubetten kostet 500 bytes.
 *
 * Unterstuetzte Excape-Sequenzen:
 * \f Seitenvorschub (glcd_clear)
 * \a Alarm (sofort darstellen, glcd_drawnow
 * \b Backspace
 * \t Tab
 * \n Newline
 * \e = \033 Escape (weiterer Befehl, z.B. andere Schriftart)
 *
 * Zeichensatz Quelltext: ISO 8859-1
 *
 * Autor: Nicolas
 */


#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include "./glcd_textfun.h"
#include "./font.h"


// Schriftarten: 


extern constflash unsigned char Font48Lookup[];
extern constflash unsigned char Font58Lookup[];
extern constflash unsigned char Font36Lookup[];
extern constflash unsigned char Font8x8Lookup[];



/* Vereinbarungen fuer die Zeichensatztabelle */
#define FONT_OFFSET_FONTWIDTH 0
#define FONT_OFFSET_CODEPAGE  1
#define FONT_OFFSET_FIRSTCHAR 2



/* Zustand der Textausgabe (inkl. Default-Werte) */
struct 
{
    constflash unsigned char * pfont; // Schriftart; Zeiger auf Schrifttabelle
    uint8_t               background; // Hintergrundpixel
    uint8_t               downshift;
    bool isFixedWidth;               
} Textcfg = 
{
    .pfont        = Font58Lookup,
    .background   = 0x00,
    .downshift    = 0,
    .isFixedWidth = true,                                                   
};



static void backspace(void);          /* Ein Zeichen zurueckspringen */
static void tab(void);                /* Tabulator */
static void newline(void);            /* Zeilenvorschub */
static bool glcd_checkescape(char c); /* Escape pruefen, Fonteinstellungen */
static uint8_t glcd_codepage(char c, char cp);  /* Codepage interpretieren */



/* Zeichen ausgeben */
void glcd_putchar(char ch)
{
    // Hilfsgroessen fuer die Schriftdarstellung. Die Hilfsgroessen koennen
    // direkt am Anfang aus dem Struct "Textcfg" ausgelesen werden, da bei einer
    // Umschaltung der Schrifteigenschafen die Funktion ohne Darstellung sofort
    // verlassen wird.
    constflash unsigned char *pfont = Textcfg.pfont; 
    uint8_t fontWidth;     // Schriftbreite in Pixeln ohne Leerspalte
    uint8_t vline;         // Pixelmuster einer senkrechten Reihe
    uint8_t c;             // Zeichen-Index
    uint8_t i;             // Zaehlindex    
    
    
    // Pruefen, ob Escape-Sequenz vorliegt. Wenn ja, beenden
    if (glcd_checkescape(ch)) 
    {
        return;
    }
    
    // Normale Zeichen: Font-Code ermitteln
    c = glcd_codepage(ch,*(pfont+FONT_OFFSET_CODEPAGE ));
       
        
    fontWidth = *(pfont+FONT_OFFSET_FONTWIDTH );
    pfont += c*fontWidth+FONT_OFFSET_FIRSTCHAR;
        
    if (Textcfg.isFixedWidth) 
    {
        // Feste Schriftbreite
        for (i=fontWidth; i>0; i--) 
        {
            vline = (*pfont++)<<Textcfg.downshift;
            glcd_putdat(Textcfg.background^vline);
        }
    } 
    else 
    {
        // Variable Schriftbreite
        for (i=fontWidth; i>1; i--) 
        {
            vline = (*pfont++)<<Textcfg.downshift;
            if (vline) glcd_putdat(Textcfg.background^vline);
        }
        vline = (*pfont)<<Textcfg.downshift;   // Letzte Spalte darstellen, auch 
        glcd_putdat(Textcfg.background^vline); // wenn sie leer ist 
    }
}



/* Strings auf LCD ausgeben */
void glcd_putstr(char *zeichenkette) 
{
    while (*zeichenkette) 
    {
        glcd_putchar(*zeichenkette);
        zeichenkette++;
    }
}



/* Strings aus Flash auf LCD ausgeben */
void glcd_putstr_P(constflash char *zeichenkette) 
{
    char temp = *zeichenkette; // Hauptsache nicht Null
    uint8_t i = 1;

    while (temp) 
    {
        glcd_putchar(temp);
        temp = *(zeichenkette+i);
        i++;
    }
}



/* Alle Zeichen zum Testen wiedergeben */
void glcd_test_chars(void) 
{
    uint8_t i = 0;
    do 
    {
        glcd_putchar(i);
    } while ( i++ != 128 );
  
} 




/* ************************************************************************** */
/*  Hilfsfunktionen                                                           */
/* ************************************************************************** */


/* Ein Zeichen zurueckspringen */
static void backspace(void)
{
    uint8_t fontWidth;
    uint16_t tmp;
    fontWidth = *(Textcfg.pfont + FONT_OFFSET_FONTWIDTH);
    if ( Textcfg.isFixedWidth==0 ) fontWidth = 1;
    tmp = glcd_getpos();
    if (tmp) 
    {
        tmp-=fontWidth;
    } 
    else 
    {
        tmp = GLCD_END;
    };
    glcd_setpos(tmp);
}


/* Tabulator */
static void tab(void)
{
    uint16_t tmp = glcd_getpos();
    if ( 0 == (tmp%TABWIDTH) ) return;
    
    tmp = tmp - (tmp%TABWIDTH) + TABWIDTH;
    tmp %= GLCD_END;
    glcd_setpos(tmp);
}


/* Zeilenvorschub */
static void newline(void)
{
    uint16_t tmp = glcd_getpos();
    tmp = tmp - (tmp%LINEWIDTH) + LINEWIDTH;
    tmp %= GLCD_END;
    glcd_setpos(tmp);

}


/* Auf Escape-Sequenzen pruefen; wenn Escape-Sequenz, wird
 *  1. diese ausgefuehrt und 
 *  2. true zurueckgeliefert, damit die Hauptfunktion beendet werden kann. */
static bool glcd_checkescape(char c)
{
    static uint8_t iscommand = 0; // printable character oder Kommando
    bool retval = false;
    
    // Falls vorangegangenes Zeichen "Escape": Kommando interpretieren
    if (iscommand) 
    {
        // Escape-Zeichen interpretieren
        switch (c) 
        {
            case 'f': // feste Schriftweite
                Textcfg.isFixedWidth = 1;
                break;
            case 'v': // variable Schriftweite
                Textcfg.isFixedWidth = 0;
                break;
            case 'i': // Invertieren
                Textcfg.background = ~Textcfg.background;
                break;
            case 's': // tiefgestellt, kleiner Zeichensatz
                Textcfg.pfont = Font36Lookup;
                Textcfg.downshift = 1;
                break;
            case 'n': // normaler Zeichensatz
                Textcfg.pfont = Font58Lookup;
                Textcfg.downshift = 0;
                Textcfg.background = 0x00U;
                Textcfg.isFixedWidth = 1;
                break;
            case 'b': // fetter Zeichensatz
                Textcfg.pfont = Font8x8Lookup;
                Textcfg.downshift = 0;
                break;
            case 'd': // extrem gequetschter Zeichensatz
                Textcfg.pfont = Font48Lookup;
                Textcfg.downshift = 0;
                break;
        }
        iscommand = 0;
        return true;
    }    
    
    // Nach Escape-Zeichen suchen:
    switch(c) 
    {
        case '\033': // = '\e'
            // Escape: Das folgende Zeichen hat Kommandofunktion
            iscommand = 1;
            retval = true;
            break;
        case '\b': // Backspace
            backspace();      
            retval = true;
            break;
        case '\t': // Tabulator
            tab();
            retval = true;
            break;
        case '\n': // newline
            newline();
            retval = true;
            break;
        case '\f': // Formfeed
            glcd_clear();
            retval = true;
            break;
        case '\a': // Alert
            glcd_drawnow();
            retval = true;
            break;
    }
    
    return retval;
}      


/* Codepage interpretieren, d.h. aus Zeichen Index im Zeichensatz erzeugen */
/* Bei MinGW ist char signed, deswergen sind Umlaute kleiner als Null. */
static uint8_t glcd_codepage(char c, char cp)
{
	uint8_t code = c;
    switch (cp) 
    {
        case CP_COMPACT: // Codepage selbstgemalte Fonts
        

			if (0x20<=c && c<=0x7D ) 
            {
				// Normaler ASCII-Bereich
				code = c;
			}
			else 
            {
				// Sonderzeichen
                switch(c) 
                {
                    case '': code=1;     break;// 
                    case '': code=2;     break;// 
                    case '': code=3;     break;// 
                    case '': code=4;     break;// 
                    case '': code=5;     break;// 
                    case '': code=6;     break;// 
                    case '': code=7;     break;// Gradzeichen
                    case '': code=0x0e;  break;// 
                    case '': code=0x0f;  break; // mu
                    case 0x10: code=0x10; break; // > mit Fuellung
                    case 0x11: code=0x11; break; // < mit Fuellung
                    case 0x12: code=0x12; break; // Schachbrett, echter Code unbekannt
                    case 0x13: code=0x13; break; // Pausenzeichen, echter Code unbekannt
                    default:  code=0;  // Falschzeichen: Box
                }
            }
            break;
            
        case CP_MSDOS: // Codepage 850
        	if (0x20<=c && c<=0x7A )
            {
        		// Normaler ASCII-Bereich
        		code = c;
        	}
        	else 
            {
        		// Sonderzeichen
        		// Benutzte Sonderzeichen werden von ISO 8859-1 auf Codepage 850
        		// gemappt, alle anderen Zeichen einfach durchgereicht
                switch(c) 
                {
                    case '': code=0x8e; break; // 
                    case '': code=0x99; break; // 
                    case '': code=0x9a; break; // 
                    case '': code=0x84; break; // 
                    case '': code=0x94; break; // 
                    case '': code=0x81; break; // 
                    case '': code=0xe1; break; // 
                    case '': code=0xe6; break; // mu
                    case 0x10: code=0x10; break; // > mit Fuellung
                    case 0x11: code=0x11; break; // < mit Fuellung
                    case 0x12: code=0xB1; break; // Schachbrett, echter Code unbekannt
                    default: code=c;
                }
            }
    }
    return code;
}

