/*
 * i2c_generic.h
 *
 * I2C generische Routinen (AVR+STM32)
 * "generisch" heisst in diesem Fall nicht Plattformunabhaengigkeit, sondern
 * dass alle I2C-Implementierungen auf diesen Grundstock zurueckgreifen.
 *
 * Fuer die Timeouts wird erwartet, dass der Timeout-Timer (setTimeout und
 * pollTimeout) laeuft.
 *
 * Erzeugt: 06.04.2014
 *  Author: DL1DOW
 */

#ifndef I2C_GENERIC_H
    #define I2C_GENERIC_H

    /* *** Einstellungen  *************************************************** */

    // Taktfrequenz der clock-Leitung
    // beim TMC222 funktionieren maximal 250000 zuverlaessig, 280000
    // funktioniert oft
    #define I2C1_CLOCK 250000
    #define I2C2_CLOCK I2C1_CLOCK

    // Tastgrad (nur auf STM32 einstellbar):
    //   I2C_DUTY_CYCLE 0: 50:50
    //   I2C_DUTY_CYCLE 1: 16:9
    #define I2C_DUTY_CYCLE 1

    #define I2C_BITBANGING_DELAY_US 10 // 10 s ergibt 50 kHz

    /* Timeouts fuer jede I2C-Operation in Millisekunden */
    #define I2C_SCANTIMEOUT 200U
    #define I2C_NORMALTIMEOUT 800U
    #define I2C_BUSYTIMEOUT 5000U
    #define I2C_RESETTIMEOUT 800U

    /* *** ab hier keine Einstellungen mehr ********************************* */


    #include <stdint.h>
    #include <stdbool.h>

    #ifdef AVR
        // I2C-Konstanten wie beim STM32 fuer den AVR definieren. Da der AVR nur
        // einen I2C hat wird diese Einstellung ignoriert
        #define  I2C1 0x00
        #define  I2C2 0x00
        typedef const uint8_t I2C_TypeDef;

    #elif defined(STM32F10x)
        #include "stm32f10x_gpio.h"
        #include "stm32f10x_rcc.h"
        #include "stm32f10x_i2c.h"

    #elif defined(STM32F4xx)
        #include "stm32f4xx_gpio.h"
        #warning not implemnted and tested for STM32F4xx
       	
    #elif defined _MSC_VER || defined __MINGW64__ || defined __linux__
        // I2C-Konstanten wie beim STM32 definieren. Nur Dummys.
        #define  I2C1 0x00
        #define  I2C2 0x00
        typedef const uint8_t I2C_TypeDef;  
        
        
    #else
        #error "unknown MCU"
    #endif

    enum {i2c_ok,       //0
        i2c_noStart,      //1
        i2c_startTimeout, //2
        i2c_noAck,        //3
        i2c_ackTimeout,   //4
        i2c_nAckTimeout,  //5
        i2c_otherTimeout, //6
        i2c_busyTimeout,  //7
        i2c_off,          //8
        i2c_stop,         //9
        i2cError          //A
    };

    // Generische Funktionen (hier implementiert)
    uint_fast8_t I2C_init(I2C_TypeDef* I2Cx);
    uint_fast8_t I2C_reset(I2C_TypeDef* I2Cx);


    // Plattform- und implementierungsabhaengige Funktionen (muessen extern
    // implementiert sein)
    enum stopbit_e {noStop,stop};
    uint_fast8_t I2C_Sendbuffer(I2C_TypeDef *I2Cx, uint8_t slaveaddr,
       uint8_t* pBuffer, uint8_t nBuffer, enum stopbit_e stopbit);

    uint_fast8_t I2C_Receivebuffer(I2C_TypeDef *I2Cx, uint8_t slaveaddr,
       uint8_t* pBuffer, uint8_t nByte);

    uint_fast8_t I2C_ping(I2C_TypeDef *I2Cx, uint8_t slaveaddr);

#endif
