/*
 * i2c_simple_avr.c
 *
 * Einfache Variante von I2C-Routinen
 * als bung zur Inbetriebnahme
 *
 * Timeouts benoetigen externen Timer uint_least24_t glu_millis (sonst wird z.B.
 * der Scan steckenbleiben)
 * 
 * 08.02.14 DL1DOW
 */


#ifndef AVR
    #error This source is designed for AVR only
#endif


//FIXME: I2C-Reset enthaelt noch kein freitakten

#include <stdint.h>
#include <avr/io.h>
#include <util/twi.h>
#include <util/delay.h>
#include "../../ioMapping.h"
#include "../src_misc/clock.h" // timeout
#include "i2c_simple.h"



// Pins vom I2C sind am AVR fest
#define I2C_CLK DD0
#define I2C_DATA DD1


/* Busy-Flag abfragen */
static inline uint8_t I2C_isBusy(void)
{
    return TWCR & (1<<TWIE);
}

#define checkforerror(s,l) if(s) {s |= (l<<4); goto error;}


/* I2C-Bus inititalisieren (ATmega32) */
uint8_t I2C_init(I2C_TypeDef* I2Cx) 
{
    (void) I2Cx; // warning unterdruecken
  
    /*set bus speed (ca. 100kHz bei 16MHz) */
    TWSR |= (0<<TWPS1)|(0<<TWPS0);
    //TWBR = 0x72;
    #define TWBR_REG (F_CPU/I2C1_CLOCK/2 - 8)
    #if TWBR_REG < 20
        TWBR = 20;
    #elif TWBR_REG > 128
        #error "invalid i2c rate"
    #else
        TWBR = TWBR_REG;
    #endif
    
    /* Prepare Pins */
    I2C_DDR &= ~((1<<I2C_SCL)|(1<<I2C_SDA));
    I2C_PORT |= (1<<I2C_SCL) | (1<<I2C_SDA); // Pull-Ups are external
  
    return i2c_ok;
}


/* Reset ohne freitakten des Busses */
uint8_t I2C_reset(I2C_TypeDef* I2Cx) 
{
    (void) I2Cx; // warning unterdruecken
    timeoutstor_t timeout;
  
    //; TWI STOP
    TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWSTO);
  
    //; TWI aus//
    TWCR = (0<<TWEN);
  
    setTimeout(&timeout,I2C_BUSYTIMEOUT);
    while(!pollTimeout(&timeout));
    I2C_init(I2Cx);

    //; Initialisierung
    return 0;
}


/* Ueberpruefen, ob der Bus gerade frei ist */
uint8_t I2C_waitbusy(I2C_TypeDef* I2Cx) 
{
    (void) I2Cx; // warning unterdruecken
  	timeoutstor_t timeout;
  	
    setTimeout(&timeout,I2C_BUSYTIMEOUT);
    while(!pollTimeout(&timeout)) {
   	    if (  (I2C_PIN & (1<<I2C_CLK) ) && !(TWCR & (1<<TWIE))  ) {
             return i2c_ok;
   	    }
    }  
    return i2c_busyTimeout;
}


/* Start condition und R/W-Bit senden
 * I2Cx     : I2C1 oder I2C2, wird ignoriert
 * slaveaddr: 7-Bit-Adresse linksbuendig
 * direction: I2C_Direction_Tranmitter (Master+Senden) oder
 *            I2C_Direction_Receiver (Master+Empfangen)
 *
 * return   : i2c_ok;          Start-Bedingung erfolgreich gesendet
 *            i2c_ackTimeout;  kein ACK empfangen
 *            i2c_noAck;       NACK explizit empfangen
 *            i2c_startTimeout;Start-Bedingung konnte nicht gesendet werden */
 uint_fast8_t I2C_start(I2C_TypeDef* I2Cx, uint8_t slaveaddr, uint8_t direction)
 {
     (void) I2Cx; // warning unterdruecken
     timeoutstor_t timeout;
     
     /* TWI aktivieren und Start-Condition ausloesen
     * warten auf START-Condition oder timeout
     * Wenn keine Start-Condition gesendet wurde abbrechen */
     TWCR = ((1<<TWINT)|(1<<TWSTA)|(1<<TWEN));
     setTimeout(&timeout,I2C_NORMALTIMEOUT);
     while( !(TWCR & (1<<TWINT)) ) {
         if(pollTimeout(&timeout)) {
             return i2c_startTimeout;
         }
     }
     
     if ( ( (TWSR & 0xF8) != TW_START) &&  ( (TWSR & 0xF8) != TW_REP_START)  ) {
         return i2c_noStart;
     }
     
     /* Adresse senden und auf Ack oder Timeout warten */
     slaveaddr &= 0xFE;
     if (direction==I2C_Direction_Transmitter)
         TWDR = slaveaddr | TW_WRITE;            // Adresse + Schreibbit ins Datenregister
     else if (direction==I2C_Direction_Receiver)
         TWDR = slaveaddr | TW_READ;             // Adresse + Lesebit ins Datenregister
     TWCR = ((1<<TWINT)|(1<<TWEN));              // senden
     
     setTimeout(&timeout,I2C_NORMALTIMEOUT);     // Timeout-Zaehler aufziehen
     while(!(TWCR & (1<<TWINT)) ) {              // warten auf ACK oder NACK
         if(pollTimeout(&timeout))
             return i2c_ackTimeout;              // Bei Timeout abbrechen
     }         
     if (  ((TWSR & 0xF8) != TW_MT_SLA_ACK) && ((TWSR & 0xF8) != TW_MR_SLA_ACK)  ) {       
         return i2c_noAck;                       // Bei nAck auch abbrechen
     }
     return i2c_ok;
 }
  

  
uint8_t I2C_write(I2C_TypeDef* I2Cx, uint8_t data) 
{
    (void) I2Cx; // warning unterdruecken
    timeoutstor_t timeout;
  
    TWDR = data;                                   // Byte in Datenregister
    TWCR = ((1<<TWINT)|(1<<TWEN));                 // senden
    setTimeout(&timeout,I2C_NORMALTIMEOUT);        // Timeout-Zaehler aufziehen
    while(!(TWCR & (1<<TWINT)) ) {                 // warten auf ACK oder NACK
        if(pollTimeout(&timeout)) {
            return i2c_ackTimeout;                 // Bei Timeout abbrechen
        }
    }  
    if((TWSR & 0xF8) != TW_MT_DATA_ACK) {          // Wenn Slave nicht akzeptiert abbrechen
        return i2c_noAck;
    }
    return i2c_ok;     
} 
 
  
uint8_t I2C_read_ack(I2C_TypeDef* I2Cx,uint8_t *data) 
{
    (void) I2Cx; // warning unterdruecken
    timeoutstor_t timeout;
  
    TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWEA);
    setTimeout(&timeout,I2C_NORMALTIMEOUT);
    while (!(TWCR & (1<<TWINT)) ) { // warten auf Empfang
        if(pollTimeout(&timeout)) {
            return i2c_otherTimeout;                  // Bei Timeout abbrechen
        }    
    }

    *data = TWDR;
    return i2c_ok;  
}



uint8_t I2C_read_nack(I2C_TypeDef* I2Cx,uint8_t *data)  
{
    (void) I2Cx; // warning unterdruecken
    timeoutstor_t timeout;
  
    TWCR = (1<<TWINT)|(1<<TWEN);
    setTimeout(&timeout,I2C_NORMALTIMEOUT);
    while (!(TWCR & (1<<TWINT)) ) { // warten auf Empfang
        if(pollTimeout(&timeout)) {
            return i2c_otherTimeout;                  // Bei Timeout abbrechen
        }
    }    
    *data = TWDR;
    return i2c_ok;
}  
  
  
uint8_t I2C_stop(I2C_TypeDef* I2Cx) 
{
    (void) I2Cx; // warning unterdruecken
    
    TWCR = ((1<<TWINT)|(1<<TWSTO)|(1<<TWEN)); // STOP condition
  
    // Warten bis STOP-condition gesendet und Bus losgelassen
    while (TWCR & (1<<TWINT));
  
    return i2c_stop;
}  
  
  

/* Ganzen Puffer ueber I2C lesen
 * I2Cx     : I2C1 oder I2C2 (bei AVR egal)
 * slaveaddr: I2C-Slave-Adresse
 * *pBuffer : Puffer passender Groesse
 * nBuffer  : Puffergroesse
 *
 * return   : Fehlerbedingung; wenn != 0, muss I2C vom Aufrufer zurueckgesetzt
 *            und STOP-Bedingung erzeugt werden
 *
 * wenn pBuffer == NULL werden nBuffer Bytes vom Bus gelesen und verworfen.
 * wenn nBuffer == NULL wird nur Stopbedingung gesendet */
uint_fast8_t
I2C_Receivebuffer(I2C_TypeDef *I2Cx, uint8_t slaveaddr,
		uint8_t* pBuffer, uint8_t nBuffer)
{
	uint8_t status, i, byte;

	if (nBuffer==0) {
		I2C_stop(I2Cx);
		return 0;
	}

	status = I2C_start(I2Cx,slaveaddr, I2C_Direction_Receiver);
	checkforerror(status,5);

	for(i=0;i<nBuffer-1;i++) {
		status = I2C_read_ack(I2Cx,&byte);
		checkforerror(status,6);
		if(pBuffer) {
			pBuffer[i] = byte;
		}
	}

	status = I2C_read_nack(I2Cx,&byte);
	checkforerror(status,7);
	if(pBuffer) {
		pBuffer[nBuffer-1] = byte;
	}

	error:
	I2C_stop(I2Cx);
	return status;
}


/* Ganzen Puffer ueber I2C senden
 * I2Cx     : I2C1 oder I2C2 (bei AVR egal)
 * slaveaddr: I2C-Slave-Adresse
 * *pBuffer : Puffer passender Groesse
 * nBuffer  : Puffergroesse
 * stop     : 1: Stop-Bedingung wird gesendet (normales Senden an Slave)
 *            0: Stop-Bedingung wird nicht gesendet (Antwort von Slave
 *            vorbereiten)
 *
 * return   : Fehlerbedingung; wenn != 0, muss I2C vom Aufrufer zurueckgesetzt
 *            und STOP-Bedingung erzeugt werden */
uint_fast8_t
I2C_Sendbuffer(I2C_TypeDef *I2Cx, uint8_t slaveaddr,
		uint8_t* pBuffer, uint8_t nBuffer, enum stopbit_e stopbit)
{
	uint_fast8_t status, i;

	status = I2C_waitbusy(I2Cx);
	checkforerror(status,1);

	status = I2C_start(I2Cx,slaveaddr, I2C_Direction_Transmitter);
	checkforerror(status,2);

	for(i=0;i<nBuffer;i++) {
	    status = I2C_write(I2Cx,pBuffer[i]);
	    checkforerror(status,4);
	}

	switch (stopbit) {
	case stop:	
		I2C_stop(I2Cx);
		break;
	case noStop:
		;
	}	

		  
    error:
	return status;
}


/* I2C-Adresse kurz anpingen um zu sehen, ob Geraet vorhanden
 *
 * return: i2c_ok          : Device vorhanden und hat mit ACK bestaetigt
 *         i2c_busyTimeout : Bus nicht frei
 *         i2c_otherTimeout: Unerwarteter Fehler
 *         i2c_ackTimeout  : NACK, Device nicht vorhanden */
uint_fast8_t
I2C_ping(I2C_TypeDef *I2Cx, uint8_t slaveaddr)
{
	uint_fast8_t status;

	// i2c_ok oder i2c_busyTimeout
	status = I2C_waitbusy(I2Cx);
	if(status) {
	    status |= (2<<4);
	    goto error;
	}

	// Startbedingung
	// i2c_ok           : ACK empfangen, Teilnehmer vorhanden
	// i2c_ackTimeout   : kein ACK empfangen, keine Antwort vom Teilnehmer
	// i2c_otherTimeout : Start-Bedingung konnte nicht gesendet werden.
	// Senderfehler.
	status = I2C_start(I2Cx,slaveaddr,I2C_Direction_Transmitter);
	I2C_stop(I2Cx);

	error:
	return status;
}



