/*
 * i2c_simple_sdl.c
 *
 * Einfache Variante von I2C-Routinen
 *
 * Primitiver Stub, der immer OK zurueckgibt
 *
 * 10.04.15 DL1DOW
 */
 
#include <stdio.h>
#include "i2c_simple.h"
#include "i2c_generic.h"
#include "../../main.h"        // #define I2C_simple
#include "../src_misc/clock.h" // timeout


#define checkforerror(s,l) if(s) {s |= (l<<4); goto error;}



/* Busy-Status abwarten, muss vor jedem Start ausgefuehrt werden */
uint_fast8_t I2C_waitbusy(I2C_TypeDef* I2Cx)
{
    printf("I2C_waitbusy 0x%x\n",(int) (intptr_t) I2Cx);    
	return i2c_ok;
}


/* Start condition und R/W-Bit senden
 * I2Cx     : I2C1 oder I2C2
 * address  : 7-Bit-Adresse linksbuendig
 * direction: I2C_Direction_Tranmitter (Master+Senden) oder
 *            I2C_Direction_Receiver (Master+Empfangen) */
uint_fast8_t I2C_start(I2C_TypeDef* I2Cx, uint8_t address, uint8_t direction)
{
	printf("I2C_start 0x%x -> %x,%i \n",(int) (intptr_t) I2Cx,address,direction);
	return i2c_ok;
}


/* Ein Byte an Slave senden */
uint_fast8_t I2C_write(I2C_TypeDef* I2Cx, uint8_t data)
{
	printf("I2C_write 0x%x -> %x\n",(int) (intptr_t) I2Cx,data);
	return i2c_ok;
}


/* Ein Byte vom Slave lesen und mit ACK besttigen = ein weiteres Byte
 * anfordern */
uint_fast8_t I2C_read_ack(I2C_TypeDef* I2Cx, uint8_t *data)
{
	*data = 0;
	printf("I2C_read_ack 0x%x \n",(int) (intptr_t) I2Cx);
	return i2c_ok;
}


/* Ein Byte vom Slave lesen und mit NACK besttigen = kein weiteres Byte
 * anfordern */
uint_fast8_t I2C_read_nack(I2C_TypeDef* I2Cx, uint8_t *data)
{
	*data = 0;
	printf("I2C_read_nack 0x%x \n",(int) (intptr_t) I2Cx);
	return i2c_ok;
}


/* Sendet STOP condition und gibt den Bus frei */
uint_fast8_t I2C_stop(I2C_TypeDef* I2Cx)
{
	printf("I2C_stop 0x%x \n",(int) (intptr_t) I2Cx);
	return i2c_stop;
}




/* Ganzen Puffer ueber I2C lesen
 * I2Cx     : I2C1 oder I2C2 (bei AVR egal)
 * slaveaddr: I2C-Slave-Adresse
 * *pBuffer : Puffer passender Groesse
 * nBuffer  : Puffergroesse
 *
 * return   : Fehlerbedingung; wenn != 0, muss I2C vom Aufrufer zurueckgesetzt
 *            und STOP-Bedingung erzeugt werden
 *
 * wenn pBuffer == NULL werden nBuffer Bytes vom Bus gelesen und verworfen.
 * wenn nBuffer == NULL wird nur Stopbedingung gesendet */
uint_fast8_t
I2C_Receivebuffer(I2C_TypeDef *I2Cx, uint8_t slaveaddr,
		uint8_t* pBuffer, uint8_t nBuffer)
{
	uint8_t status, i, byte;

	if (nBuffer==0) {
		I2C_stop(I2Cx);
		return 0;
	}

	status = I2C_start(I2Cx,slaveaddr, I2C_Direction_Receiver);
	checkforerror(status,5);

	for(i=0;i<nBuffer-1;i++) {
		status = I2C_read_ack(I2Cx,&byte);
		checkforerror(status,6);
		if(pBuffer) {
			pBuffer[i] = byte;
		}
	}

	status = I2C_read_nack(I2Cx,&byte);
	checkforerror(status,7);
	if(pBuffer) {
		pBuffer[nBuffer-1] = byte;
	}

	error:
	I2C_stop(I2Cx);
	return status;
}


/* Ganzen Puffer ueber I2C senden
 * I2Cx     : I2C1 oder I2C2 (bei AVR egal)
 * slaveaddr: I2C-Slave-Adresse
 * *pBuffer : Puffer passender Groesse
 * nBuffer  : Puffergroesse
 * stop     : 1: Stop-Bedingung wird gesendet (normales Senden an Slave)
 *            0: Stop-Bedingung wird nicht gesendet (Antwort von Slave
 *            vorbereiten)
 *
 * return   : Fehlerbedingung; wenn != 0, muss I2C vom Aufrufer zurueckgesetzt
 *            und STOP-Bedingung erzeugt werden  */
uint_fast8_t
I2C_Sendbuffer(I2C_TypeDef *I2Cx, uint8_t slaveaddr,
		uint8_t* pBuffer, uint8_t nBuffer, enum stopbit_e stopbit )
{
	  uint_fast8_t status, i;

	  status = I2C_waitbusy(I2Cx);
	  checkforerror(status,1);

	  status = I2C_start(I2Cx,slaveaddr, I2C_Direction_Transmitter);
	  checkforerror(status,2);

	  for(i=0;i<nBuffer;i++) {
	    status = I2C_write(I2Cx,pBuffer[i]);
	    checkforerror(status,4);
	  }

	  switch (stopbit) {
	  case stop:
		  I2C_stop(I2Cx);
		  break;
	  case noStop:
		  ;
	  }

	  error:
	  return status;
}


/* I2C-Adresse kurz anpingen um zu sehen, ob Geraet vorhanden
 *
* return: i2c_ok          : Device vorhanden und hat mit ACK bestaetigt
 *         i2c_busyTimeout : Bus nicht frei
 *         i2c_otherTimeout: Unerwarteter Fehler
 *         i2c_ackTimeout  : NACK, Device nicht vorhanden */
uint_fast8_t
I2C_ping(I2C_TypeDef *I2Cx, uint8_t slaveaddr)
{
	uint_fast8_t status;

	// i2c_ok oder i2c_busyTimeout
	status = I2C_waitbusy(I2Cx);
	if(status) {
	    status |= (2<<4);
	    goto error;
	}

	// Startbedingung
	// i2c_ok           : ACK empfangen, Teilnehmer vorhanden
	// i2c_ackTimeout   : kein ACK empfangen, keine Antwort vom Teilnehmer
	// i2c_otherTimeout : Start-Bedingung konnte nicht gesendet werden.
	//                    Senderfehler.
	status = I2C_start(I2Cx,slaveaddr,I2C_Direction_Transmitter);
	I2C_stop(I2Cx);

	error:
	return status;
}


