/*
 * menu.c
 *
 * - Darstellung eines Menues
 * - Darstellung von "Popup-Fenstern"
 *
 * Created: 10.07.2013 18:15:08
 *  Author: Nicolas
 */



#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include "../src_base/assert.h"
#include "../src_base/intmath.h"
#include "../src_avrcom/delay.h"
#include "../src_glcd/glcd_generic.h"
#include "../src_glcd/glcd_textfun.h"
#include "../src_glcd/glcd_gfxfun.h"
#include "../src_encoder/encoder_panasonic.h"
#include "../src_encoder/key.h"
#include "menu.h"





// Einstellungen Menudarstellung
#define LINEWIDTH 128U // In Pixel
#define DISPHEIGHT 8U  // In Zeilen


#ifdef _MSC_VER // Windows
  #define UPDATE_DISPLAY_EVERY 1
#else
  #define UPDATE_DISPLAY_EVERY 20
#endif



#define MIN(X,Y) ((X) < (Y) ? (X) : (Y))



/* ************************************************************************************ */
/* Menu: Darstellungsfunktionen                                                         */
/* ************************************************************************************ */
/* -Scrollbalken
 * -Menutexte
 * -Cursor
 */


/* Darstellung eines senkrechten Scrollbalkens am rechten Bildrand
 *  currentPos: Position im Scrollbereich
 *  nPos: Groesse des Scrollbereichts (Anzahl der Positionen)
 *  nPos = 0 -> kein Slider
 *  height: Hoehe der Scrollbar (in x8 Pixel); von unterer Displaykante */    
const uint8_t PROGMEM scrollgfx [] = 
{
    0xFF, 0x5B, 0x99, 0x5B, 0xFF,  // Pfeil hoch
    0xFF, 0x55, 0xAA, 0x55, 0xFF,  // Mittelstueck
    0xFF, 0xD9, 0x9A, 0xD9, 0xFF,  // Pfeil nach unten
    0xAB, 0x81, 0x81, 0x81, 0xAB,  // "Schieber"
};


#define offset_slidergfx 15        // Offset der Minigrafiken
#define offset_botgfx 10
#define offset_topgfx 0
#define offset_midgfx 5

typedef struct 
{
    uint8_t nMenuitem;    // Anzahl der Menuepunkte
    uint8_t nVisibleitem; // Anzahl sichtbare Menueeintraege =< Scrollbereich
    uint8_t nHeadline;    // Anzahl der Displayzeilen oberhalb Scrollbereich
} ScrollAreaData_t;


void draw_scrollbar(uint8_t currentPos,uint8_t nPos,uint8_t nHeadline) 
{

    uint8_t scrollAreaHeight = DISPHEIGHT-nHeadline;
    uint8_t gfxoffset;
    uint8_t i, j, temp;
    uint8_t vpos; // Position des Scrollbalkens (obere Kante)
    uint8_t block1, block2; // Position der Begrenzungspixel
    uint8_t bit1, bit2;     // Position der Begrenzungspixel
    
    #define pxborder 10U   // oben und unten
    #define sliderlen 8U // Nicht ohne weiteres aenderbar!
    #define pxscrollbar (scrollAreaHeight*8U-pxborder-sliderlen) // Nutzbarer Bereich 
                                                                 // Scrollbalken in px
    #define pxoffset 5U    // Abstand beweglicher Bereich nach oben
    
    // Position des Scrollbalkens (obere Kante) 
	if(nPos==1) 
    { 
        // Kein richtiger Scrollbalken
		vpos = pxoffset;
	}
	else 
    {
        vpos = (( ((32U*((uint16_t) pxscrollbar+1U) )/(nPos-1U))*(currentPos) )/32)+pxoffset;
	}

    // Position der Begrenzungspixel
    block1 = vpos/8U;
    bit1 = vpos%8U;
    block2 = (vpos+7U)/8U;
    bit2 = (vpos+7U)%8U;
    if (nPos==0U) 
    {
        block1 = scrollAreaHeight; // ausserhalb des Bereichs
        block2 = scrollAreaHeight;
    }
    
    // Senkrechter Aufbau des Scrollbalkens
    for (j = 0;j<scrollAreaHeight;j++) 
    {
        if (j==0) 
        { 
            // Oberster Block
            gfxoffset = offset_topgfx;
        }      
        else if (j==scrollAreaHeight-1) 
        {
            // Unterster Block
            gfxoffset = offset_botgfx;
        }
        else 
        {
            // Mittlere Bloecke
            gfxoffset = offset_midgfx;
        }
        
        // Grafiken fuer Scrollbalken positionieren und Slider einfuegen
        glcd_setpos((nHeadline+1U+j)*LINEWIDTH-5U);
        
        for (i=0;i<5;i++) 
        {
            temp = pgm_read_byte(&scrollgfx[i+gfxoffset]);
            if (block1 == j) 
            { 
                // Slider untere Haelfte einfuegen
                temp &= ~(0xFF << bit1);
                temp |=  (pgm_read_byte(&scrollgfx[i+offset_slidergfx]) << bit1);
            }
            if (block2 == j) 
            { 
                // Slider obere Haelfte einfuegen
                temp &= (0xFF << bit2);
                temp |=  (pgm_read_byte(&scrollgfx[i+offset_slidergfx]) >> (7-bit2));
            }
            
            glcd_putdat(temp);
        };
    }
} // Ende Scrollbalken


/* Menue-Cursor (Pfeile rechts und links) darstellen 
 * cursorPos: Absolute Position vom oberen Displayrand */
void draw_cursor(uint8_t cursorPos) 
{
    glcd_setpos((cursorPos)*LINEWIDTH);
    glcd_putdat(0x7F);
    glcd_putdat(0x7F);
    glcd_putdat(0x00);
    glcd_putchar(0x10); //>

    glcd_setpos(cursorPos*LINEWIDTH+LINEWIDTH-14U);
    glcd_putchar(0x11); //<
    glcd_putdat(0x7F);
    glcd_putdat(0x7F);
}


/* Sichtbare Menuetexte darstellen (inkl. Hervorhebung des aktuellen Texts)   */
/*   firstItem: Anfang des darstellten Bereichs (Index)
 *   actItem:   Hervorgehobenes Element (Index)
 *   height:    Hoehe des Dargestellten Bereichs in Zeilen; von unterer 
 *              Displaykante
 */
#define GCC_VERSION ( __GNUC__ * 10000 \
		+ __GNUC_MINOR__ * 100 \
		+ __GNUC_PATCHLEVEL__ )
		
#if ( (GCC_VERSION>=40800 && GCC_VERSION<40804) || GCC_VERSION==40900)
	/* WORKAROUND:
		Bug 61443
		https://gcc.gnu.org/bugzilla/show_bug.cgi?id=61443		
		
		Unter AVR-GCC 4.8.1 sorgt die direkte Verwendung von
		"menudata[item].menutext" mit "snprintf_P" fuer einen Compilerfehler 
		("unrecognizable insn"). Konkret lautete die Zeile:
	
		snprintf_P(glstr_Buf,N_TEXTBUF,menudata[item].menutext,*menudata[item].puint8);
      
		Als Workaround wird der String zuerst ins SRAM kopiert und anschliessend 
		die snprintf-Funktion genutzt (benoetigt 81 Bytes zusaetzlich). */
	#warning "This AVR-GCC version needs workaround for flash strings"
	#define WORKAROUND_STRING
#endif


void draw_menutext(constflash menudata_t menudata[],\
    uint8_t firstItem,uint8_t selectedItem,uint8_t nItem, uint8_t nHeadline) 
{
    uint_fast8_t i, item;
	#ifdef WORKAROUND_STRING
		char tempstr[N_TEXTBUF];	
	#endif
  
    item = firstItem;
    for (i = 0; i<nItem ; i++) 
    {
    	uint16_t pos;

    	pos = ( (uint16_t) i+nHeadline);
    	pos *= LINEWIDTH;
    	pos += 9U;

        glcd_setpos(pos);
		
	    switch(menudata[item].datatype) 
        {
			case nothing:
				// Falltrough

			case voidFcn:
				// Menutexte fuer Funktionen werden einfach dargestellt
				strncpy_P(glstr_Buf,menudata[item].menutext,N_TEXTBUF);
				glstr_Buf[N_TEXTBUF-1] = '\0';
				break;

			case uint8:
				// Texte fuer Einstellkonstanten werden mit printf
				// interpretiert
				#ifdef WORKAROUND_STRING
					strncpy_P(tempstr,menudata[item].menutext,N_TEXTBUF);
					tempstr[N_TEXTBUF-1] = '\0';
					snprintf(glstr_Buf,N_TEXTBUF,tempstr,*menudata[item].puint8);
    			#else
					snprintf_P(glstr_Buf,N_TEXTBUF,menudata[item].menutext,\
					*menudata[item].puint8);
				#endif
				break;

			case uint16:
				// Texte fuer Einstellkonstanten werden mit printf
				// interpretiert			
				#ifdef WORKAROUND_STRING
					strncpy_P(tempstr,menudata[item].menutext,N_TEXTBUF);
					tempstr[N_TEXTBUF-1] = '\0';
					snprintf(glstr_Buf,N_TEXTBUF,tempstr,*menudata[item].puint16);
				#else
					snprintf_P(glstr_Buf,N_TEXTBUF,menudata[item].menutext,\
					*menudata[item].puint16);
				#endif			
				break;

			case int8:
				// Texte fuer Einstellkonstanten werden mit printf
				// interpretiert
				#ifdef WORKAROUND_STRING
					strncpy_P(tempstr,menudata[item].menutext,N_TEXTBUF);
					tempstr[N_TEXTBUF-1] = '\0';
					snprintf(glstr_Buf,N_TEXTBUF,tempstr,*menudata[item].pint8);
				#else							
					snprintf_P(glstr_Buf,N_TEXTBUF,menudata[item].menutext,\
						*menudata[item].pint8);
				#endif
				break;

			case int16:
				// Texte fuer Einstellkonstanten werden mit printf
				// interpretiert
				#ifdef WORKAROUND_STRING
					strncpy_P(tempstr,menudata[item].menutext,N_TEXTBUF);
					tempstr[N_TEXTBUF-1] = '\0';
					snprintf(glstr_Buf,N_TEXTBUF,tempstr,*menudata[item].puint16);
				#else			
					snprintf_P(glstr_Buf,N_TEXTBUF,menudata[item].menutext,\
						*menudata[item].pint16);
				#endif
				break;
		}

           
        // Normale und hervorgehobene Menuepunkte
        if (selectedItem == item) glcd_putstr_P(PSTR("\033i")); // Hervorgehoben
        glcd_putstr(glstr_Buf);
        if (selectedItem == item) glcd_putstr_P(PSTR("\033n")); // Hervorhebung
		                                                      // zuende
    
        item++;
    }
}


/* Alles zusammen darstellen (Scroll Area)                                   */
void draw_menu(constflash menudata_t menudata[],\
               uint8_t selectedItem, 
               uint8_t scroll_pos, 
               ScrollAreaData_t Sad)
{
    
    
    draw_menutext(menudata,scroll_pos,selectedItem,Sad.nVisibleitem,Sad.nHeadline);
    draw_cursor(selectedItem-scroll_pos+Sad.nHeadline);
          
    //draw_scrollbar(selection,nMenuEntries,scrollHeight); // Eintragweise
    draw_scrollbar(scroll_pos,Sad.nMenuitem-Sad.nVisibleitem+1,Sad.nHeadline); // Seitenweise
    // draw_scrollbar(scroll_pos,0,scrollHeight); // ohne Scrollen

    #ifdef GLCD_FRAMEBUFFER
        // Display-Inhalt darstelen
        glcd_drawnow();
    #else
        // Cursor zurueck auf Nullposition, damit glcd_clear von dort
        // ausgefuehrt wird. Ansonsten werden Bildteile unterschiedlich
        // lange darstellt und flackern
        glcd_setpos(0);
    #endif  
}


/* Falls Kopfzeile in Textform vorliegt, diese darstellen, Kopf-Funktion,    */
/* falls vorhanden, ausfuehren                                               */
void execute_headline(constflash menuconfig_t *menuconfig)
{
    voidFcn_t headfkt;

    if (menuconfig->pheadfkt) 
    {
        headfkt = menuconfig->pheadfkt;
        headfkt();
    }  
  
    if(menuconfig->headtext) 
    {
        printf_P(menuconfig->headtext);
    }
}


/* Menuezeile ausfuehren:
 * - Funktionen werden ausgefuehrt
 * - Variablen werden mit dem Encoder bearbeitet
 *
 * Menuitem: Menuezeile
 * enc_diff: EncoderInkrement fuer Variablenbearbeitung
 * return:   Weitere Aktion in Menue-Hauptschleife ausloesen */
typedef enum 
{
    doResetEncoder,
    doNothing,
    doReturn
} ctrl_caller_t;

ctrl_caller_t dealWithMenuItem( menudata_t Menuitem, int_fast8_t enc_diff )
{
    ctrl_caller_t callerAction; // Rueckgabewert
    
    switch (Menuitem.datatype) 
    {
        case nothing: // Kann ignoriert werden
            callerAction = doResetEncoder;
            break;
        case voidFcn: // Funktionsausfuehrung oder Menue-Beenden initiieren
            {
                voidFcn_t nextfkt = Menuitem.pvoidFcn;
                if (nextfkt==NULL)
                {
                    callerAction = doReturn;
                }                    
                else 
                {
                    nextfkt();
                    callerAction = doResetEncoder;
                }                
            }
            break;
        
        case uint8: // uint8-Variable bearbeiten
            {
                uint8_t *pu8 = Menuitem.puint8;
                int_fast8_t incr = satmul_int8(enc_diff,Menuitem.varInc);
                *pu8 = limadd_uint8(*pu8,incr,Menuitem.varMin,Menuitem.varMax);
            }
            callerAction = doNothing;
            break;
        
        case uint16: // uint16-Variable bearbeiten
            {
                uint16_t *pu16 = Menuitem.puint16;
                int16_t min = Menuitem.varMin;
                if (min < 0) error("Invalid value range");
                int16_t max = Menuitem.varMax;
                if (max < 0) error("Invalid value range");
                int_fast8_t incr = satmul_int8(enc_diff,Menuitem.varInc);
                *pu16 = limadd_uint16(*pu16,incr,(uint16_t) min,(uint16_t) max);
            }
            callerAction = doNothing;
            break;
        
        case int8: // int8-Variable bearbeiten
            {
                int8_t *pi8 = Menuitem.pint8;
                int_fast8_t incr = satmul_int8(enc_diff,Menuitem.varInc);
                *pi8 = limadd_int8(*pi8,incr,Menuitem.varMin,Menuitem.varMax);
            }
            callerAction = doNothing;
            break;
        
        case int16: // int16-Variable bearbeiten
            {
                int16_t *pi16 = Menuitem.pint16;
                int_fast8_t incr = satmul_int8(enc_diff,Menuitem.varInc);
                *pi16 = limadd_int16(*pi16,incr,Menuitem.varMin,Menuitem.varMax);
            }
            callerAction = doNothing;
            break;
        default:
            error("unexpected datatype descriptor");
            callerAction = doNothing;
    } // switch datatype
    
    return callerAction;
}





/* ************************************************************************************ */
/* Menu: Auswahl- und Manipulationsbereich                                              */
/* ************************************************************************************ */
// Darstellung Menu. Dicker Haufen Spaghetti
uint_fast8_t menu_standard(constflash menudata_t menudata[],\
                           constflash menuconfig_t *menuconfig)
{
    ScrollAreaData_t Sad;  // Innerhalb eines Menues konstante Werte
    uint_fast8_t scrollPos = 0;     // Scrollposition (oberste Zeile)
    uint_fast8_t selectedItem  = 0; // Ausgewaehlter Menupunkt, Startauswahl 0
    
	int_fast8_t enc_diff = 0;
    uint_fast8_t timeslot = 0;

    // Menumodus
    enum 
    {
        select,  // Menueauswahl
        action   // Variable bearbeiten/Funktion ausfuehren
    } encoder_mode = select;
    
   
    // Menueeintraege zaehlen und konstante Menuparameter ausrechnen
    Sad.nMenuitem = 0;
    while(strlen_P(menudata[Sad.nMenuitem].menutext)) Sad.nMenuitem++;
    
    Sad.nHeadline = menuconfig->headheight; 
    Sad.nVisibleitem = MIN(Sad.nMenuitem,DISPHEIGHT-Sad.nHeadline);        

    
    // Drehgeber ruecksetzen
    encoder_get_diff();
    
    while(1) 
    {
		// Menue als Funktionsauswahl
        
		// Drehgeber auswerten (nur Aenderungen)
		enc_diff = encoder_get_diff();

        
		// Darstellung Men (Scrollbereich), nur in einem von N Zeitschlitzen
		if (timeslot==0) 
        {
			glcd_clear();
			execute_headline(menuconfig);
			draw_menu(menudata,selectedItem,scrollPos,Sad);
		} // if (timeslot==0)
        
       
		// Auf Tastendruck reagieren und Eimgabemodus umschalten
		if(key_getstroke()) 
        {
			switch(encoder_mode) 
            {
			    case select: 
			        encoder_mode = action;
				    break;
		        case action:
			        encoder_mode = select;
			}
		}

		
        // Vorgehen nach Eingabemodus
		switch (encoder_mode) 
        {
		    case select:
			    // Drehgeber waehlt Menueeintrag
                // enc_diff negativ, damit Linksdrehung nach unten geht
                selectedItem = limadd_uint8(selectedItem,-enc_diff,0,Sad.nMenuitem-1);
		
			    // Mitscrollen (mitschleppen)
			    if (selectedItem>scrollPos+Sad.nVisibleitem-1)
				    scrollPos++;
			    if (selectedItem<scrollPos)
				    scrollPos--;
		
		        break;			
			
		    case action:
                {
                    // Funktionsaufruf oder bearbeiten einer Variable
                    ctrl_caller_t cmd = dealWithMenuItem(menudata[selectedItem],enc_diff);

                    switch(cmd) {
                    case doNothing:
                        break;
                    case doResetEncoder:
                	    encoder_get_diff();
                        encoder_mode = select;
                        break;
                    case doReturn:
                        return (uint_fast8_t) selectedItem;
                    }
                } 
    	} // switch(encoder_mode)


        delay_ms(1); // Etwas verzoegern um Flackern zu vermindern
        timeslot++;
        if (timeslot==UPDATE_DISPLAY_EVERY) timeslot = 0;
        
    } // while (1)
    
    // Wird nie erreicht, da aus while(forever) nur mit return gesprungen
    // werden kann.
    return 0;
    
} // end function




/* ************************************************************************************ */
/* Usermessage                                                                          */
/* ************************************************************************************ */
/* Usermessages sind von der Struktur wie ein einfaches Menu.
 * Hauptunterschied ist, dass das Menu statisch im flash vorliegen muss, waehrend die 
 * usermessage zur Laufzeit aus Zeigern auf statische Inhalte zusammengestellt wird.
 * Es wird auch nicht gescrollt; es muessen alle Optionen auf das Display passen. */


/* Sichtbare Menuetexte darstellen (inkl. Hervorhebung des aktuellen Texts)  */
/*   firstItem: Anfang des darstellten Bereichs (Index)
 *   actItem:   Hervorgehobenes Element (Index)
 *   height:    Hoehe des Dargestellten Bereichs in Zeilen; von unterer 
 *              Displaykante
 */
static_always_inline void draw_messagetext(constflash char * const items[],\
    uint8_t firstItem,uint8_t selectedItem,ScrollAreaData_t Sad) 
{
    uint8_t i, item;
  
    item = firstItem;
    for (i = 0; i<Sad.nMenuitem; i++) 
    {
        glcd_setpos(LINEWIDTH*( (uint_fast16_t) i+Sad.nHeadline)+9U);
        strncpy_P(glstr_Buf,items[item],N_TEXTBUF);
              
        // Normale und hervorgehobene Menuepunkt
        if (selectedItem == item) glcd_putstr_P(PSTR("\033i")); // Hervorgehoben
        glcd_putstr(glstr_Buf);
        if (selectedItem == item) glcd_putstr_P(PSTR("\033n")); // Hervorhebung zuende
    
        item++;
    }
}



uint_fast8_t usermessage(usermessage_t *message) 
{
    ScrollAreaData_t Sad;           // Innerhalb eines Menues konstante Werte    
    uint_fast8_t scrollPos = 0;     // Scrollposition (oberste Zeile)
    uint_fast8_t selectedItem  = 0; // Ausgewaehlter Menupunkt, Startauswahl 0
    
	int_fast8_t enc_diff = 0;
      
    uint8_t timeslot = 0;


    // Menueeintraege zaehlen
    Sad.nMenuitem = 0;
    while(strlen_P(message->items[Sad.nMenuitem])) Sad.nMenuitem++;
      
    // Konstanten ausrechnen
    Sad.nHeadline = message->headheight;
    Sad.nVisibleitem = MIN(Sad.nMenuitem,DISPHEIGHT-Sad.nHeadline);    
      
    // Die Menuesteuerung uebernimmt den Drehgeber: Drehgeber ruecksetzen
    encoder_get_diff();
      
    while(1) 
    {
        // Menue als Funktionsauswahl
        
        // Drehgeber auswerten (nur Aenderungen)
        enc_diff = encoder_get_diff();
        
        // Limitierung Menueposition
        selectedItem = limadd_uint8(selectedItem,-enc_diff,0,Sad.nMenuitem-1);
        
        // Mitscrollen (mitschleppen)
        if (selectedItem>scrollPos+Sad.nVisibleitem-1) scrollPos++;
        if (selectedItem<scrollPos) scrollPos--;

        
        // Darstellung Auswahl
        if (timeslot==0) 
        {
            glcd_clear();
            
            // Kopfzeile darstellen.
            // Ueber den Integer errcode kann auch eine Fehler-Information dargestellt 
            // werden
            // printf ignoriert das ueberschuessige Argument, wenn es im Formatstring 
            // nicht genutzt wird. (siehe ISO/IEC 9899:TC3, 7.19.6.1)
            snprintf_P(glstr_Buf,N_TEXTBUF,message->text,message->errcode);
            glcd_putstr(glstr_Buf);

            // Piktogramm darstellen
            glcd_setpos(0);
            glcd_putbitmap_P(message->icon);
            glcd_setpos(128-16);
            glcd_putbitmap_P(message->icon);
            
            // Text darstellen
            draw_messagetext(message->items,scrollPos,selectedItem,Sad);
          
            #ifdef GLCD_FRAMEBUFFER
                // Display-Inhalt darstelen
                glcd_drawnow();
            #else
                // Cursor zurueck auf Nullposition, damit glcd_clear von dort
                // ausgefuehrt wird. Ansonsten werden Bildteile unterschiedlich
                // lange darstellt und flackern
                glcd_setpos(0);
            #endif
          
        } // if (timeslot==0)
        
        // Auf Knopfdruck reagieren: Mit Auswahlwert Funktion beenden
        if (key_getstroke()) 
        {
            // Rueckgabewert ist "selectedItem"
            break;
        }

        delay_ms(1); // Etwas verzoegern um Flackern zu vermindern
        timeslot++;
        if (timeslot==UPDATE_DISPLAY_EVERY) timeslot = 0;
    } // while (1)

    return selectedItem;
    
} // end function

