/*--------------------------------------------------------------------------------------
 * Titel     : Firmware Rotorsteuergeraet fuer FTM25/6
 *--------------------------------------------------------------------------------------
 * Funktion  : siehe ausfuehrliche Dokumentation unter
 *             http://DL1DOW.de/inhalt/selbstbau/ftmctrl/index.htm
 *
 * Schaltung : FTM25/6-Steuerplatine Rev. 1.0 ohne Modifikationen 
 *--------------------------------------------------------------------------------------
 * Hardware  : AVR ATmega32
 * Takt      : 7,3728 MHz
 * Sprache   : AVR-GCC 4.8.1
 * Datum     : 19.02.2016
 * Version   : siehe "svn_version.h"
 * Autor     : Nicolas, DL1DOW
 *--------------------------------------------------------------------------------------*/




#include "main.h"

#include "./common/src_avrcom/delay.h"
#include "./common/src_avrcom/platform.h"
#include "./common/src_avrcom/io.h"
#include "./common/src_encoder/key.h"     
#include "./common/src_encoder/encoder_panasonic.h"
#include "./common/src_glcd/glcd_textfun.h"
#include "./common/src_misc/lowlevel.h"              // jtag_disable()
#include "./common/src_misc/clock.h"                 // glu_millis
#include "./common/src_uart/uart.h"
#include "./common/src_ui/ui_infoscreen.h"

#if defined(PLATFORM_AVR)
	#include <avr/interrupt.h>
	#include <avr/eeprom.h>
    
#elif defined (PLATFORM_PC)    
    #include <stdio.h>
    #include <SDL2/SDL.h>
	#include "./common/src_mockup_sdl/mock_sdl.h"
    #warning "PLATFORM PC"
    
#endif
 
#include "./project/src_adc/adc_multiplex.h"
#include "./project/src_rotctrl/ftm_menu.h"
#include "./project/src_rotctrl/pulsecounter.h"
#include "./project/src_rotctrl/ftm_ctrl.h"
#include "./project/src_rotctrl/ftm_eeprom.h" // restorePulse()
#include "ioMapping.h"





#ifdef PLATFORM_AVR
    /* Hilfsfunktionen und -Definitionen fuer stdout und stderr */
    static int gputchar(char c, FILE *stream) 
    {
        (void) stream;
        glcd_putchar(c);
        return 0;
    }
    static FILE mystdout = FDEV_SETUP_STREAM(gputchar, NULL,_FDEV_SETUP_WRITE);
    
#endif 



/* ************************************************************************************ */
/*  Timer-Interrupt (Haupttimer, Takt 1kHz)                                             */
/* ************************************************************************************ */

// Diese Funktion schreit nach "static inline". Allerdings ist das System ausreichend 
// schnell und damit keinerlei Anlass fuer eine Optimierung vorhanden.
void task1ms(void)
{
	encoder_get();
	glu_millis++; // u.A. fuer lange und kurze Tastendruecke
    
    count_pulses();
}



/* ************************************************************************************ */
/*     Main                                                                             */
/* ************************************************************************************ */
int main(int argc, char* argv[])
{
	(void) argc;
	(void) argv;
	
    /* *** Variablendeklarationen *** */
    // -
    
    /* *** Initialisierung *** */


	#ifdef PLATFORM_AVR
        // Zielplattform: AVR
		stdout = &mystdout; // stdout fuer printf festlegen
		stderr = &mystdout; // stderr fuer fprintf festlegen
        jtag_disable();     // Pins werden fuer GLCD gebraucht
        
    #elif defined(PLATFORM_PC)
        // PC-Mockup
        sdl_init();         // Grafik initialisieren
        
    #else
        error "Target platform not supported"    
        
	#endif


    // Entprellen muss vor dem ADC-Takt initialisiert werden.
    pulsecounter_init();
    
    
    // Der ADC-Takt liefert auch den Trigger fuer die 1ms-Interrupts
    adc_init();
    adc_start();

	
    // Grafische Benutzerschnittstelle
    glcd_init();        // KS0108-Grafik-LCD
    key_init();         // Bedienknopf/Taster
    encoder_init();     // Bedienknopf/Drehgeber
    
    
    // Nur fuer Debug-Ausgaben 
    uart_init();
    
    
    // Spare-Pin/General Purpose Pin (fuer Profiling-Zwecke)
    io_setOutput(SPARE0_GPIO,SPARE0_Pin);


	// Displaybeleuchtung an
    io_setOutput(GLCD_LIGHT_GPIO,GLCD_LIGHT_Pin);
    io_setBit(GLCD_LIGHT_GPIO,GLCD_LIGHT_Pin);


	// Input-Spannungsteiler an
	io_setOutput(EN_INPUT_GPIO,EN_INPUT_Pin);
	io_setBit(EN_INPUT_GPIO,EN_INPUT_Pin);


	#ifndef PLATFORM_PC
		// Aktuellen Pulszaehler aus EEPROM wiederherstellen
		if(restorePulse() == EXIT_FAILURE)
		{
		    goto dengo;
		}
    #endif
 
    
    // Logo erlaubt schnelle Beurteilung des Kontrasts
    splashscreen();
	delay_ms(250);
    

    // Hauptschleife
    while(loop()) 
    {
        control_ftm();
        
        dengo:
		menu_main();  
    }
        
    return 0;
} // End main



/* Alle Motoren anhalten, System blockieren, Displaybeleuchtung blinken lassen */
void abort(void)
{
    io_clearBit(MOT1_EN_GPIO,MOT1_EN_Pin);
    io_clearBit(MOT2_EN_GPIO,MOT2_EN_Pin);
    
    while(1) 
    {
        io_toggleBit(GLCD_LIGHT_GPIO,GLCD_LIGHT_Pin);
        delay_ms(250);        
    }
}

