/*
 * adc_multiplex_avr.c
 *
 * ADC-Routinen fuer FTM25/6-Rotorsteuerung
 * Gemultiplext auslesen: Der ADC wird freilaufend reihum abgefragt.
 * ADC-Interrupt dient auch als Grob-Timer fuer 1 Millisekunde
 *
 * Schreibt in Glcstat
 *
 *
 * Created: 28.01.2016
 *  Author: Nicolas
 */



#include "adc_multiplex.h"
#include <avr/interrupt.h>
#include "../../common/src_avrcom/io.h"
#include "../../ioMapping.h"
#include "../../main.h"
#include "../../common/src_base/assert.h"
#include "../../project/src_rotctrl/rot_settings.h"
#include "../../project/src_rotctrl/rot_variables.h"

#ifndef AVR
    #error "unknown MCU"
#endif




// Liste der abzufragenden Kanaele
#define N_LIST 2
const uint8_t list1[] = {MOT1_ADC_MUX,MOT2_ADC_MUX}; // Darauf wird der Mux als naechstes 
                                                     // eingestellt
const uint8_t list2[] = {MOT1_ADC_MUX,MOT2_ADC_MUX}; // Daraus wird gelesen


/* Einstellungen fuer Register ADMUX
 *  - Interne Interne 2.56V-Referenz
 *  - Ergebnis rechtsbuendig ausgerichtet */
#define ADCMUXCONFIG (1<<REFS0)|(1<<REFS1)|(0 << ADLAR) 


#define ADC_PRESCALER 128UL
#define N_ADCS_PER_MILLISECOND (F_CPU/1000/14/ADC_PRESCALER)


/* Im ADC-Interrupt-Handler findet statt:
 *  - Umschaltung des Multiplex-Kanals
 *
 *  - "Signalverarbeitung" (gleitender Mittelwert). Die Ergebnisse werden in
 *     globalen Variablen hinterlegt
 *
 *  - Gleichzeitig kann der ADC-IRQ als zentraler Zeitgeber genutzt werden
 *
 *  - Loest bei Ueberstrom Not-Halt aus  */
ISR(ADC_vect)
{			     		// 4./7. ADC ready INT
	static uint8_t i = 0;

	//io_setBit(GPP_GPIO,GPP_Pin);


	i++;
	if (i==N_LIST) i=0;
	uint8_t mux = list1[i];

	uint16_t adval = ADCW;      // 5. Wert holen
	ADCSRA &= ~(1<<ADATE);      // 6. auto-trigger off
	ADMUX = ADCMUXCONFIG | mux; // Neuer Mux-Kanal
	ADCSRA |= (1<<ADATE);	 	// auto-trigger on --> Comp.-Match


	// ADC-Wert korrekt einsortieren
	switch (list2[i]) 
    {
	    case MOT1_ADC_MUX: // PA1
		    Glcstat.mot[0].adc =  (adval + 63*Glcstat.mot[0].adc)/64;
            if( Glcstat.mot[0].adc > MAX_CURRENT_DIGIT ) 
            {
                error("Overcurrent");
            }        
		    break;
            
	    case MOT2_ADC_MUX: // PA3
		    Glcstat.mot[1].adc =  (adval + 63*Glcstat.mot[1].adc)/64;
            if( Glcstat.mot[1].adc > MAX_CURRENT_DIGIT ) 
            {
                error("Overcurrent");
            }
		    break;
	}


	// Periodische Tasks (1ms)
	static uint8_t j = 0;
	j--;
	if (j==0) 
    {
		j = N_ADCS_PER_MILLISECOND-1;
		task1ms();
	}


	//io_clearBit(GPP_GPIO,GPP_Pin);

}


/* ADC initialisieren: freilaufend, 4,1kHz 
 * 7,3728MhZ/(128*14) */

void adc_init(void)
{
    /* ADCSRA (DB S. 216ff.)
     * - ADC Enable
     * - Auto Trigger 
     * - Interrupt an
     * - Prescaler 128 */
    ADCSRA =  (1<<ADEN) |  (1<<ADATE) | (1<<ADIE) | \
         (1<<ADPS2)  | (1<<ADPS1) | (1<<ADPS0);
         
    ADMUX = ADCMUXCONFIG | 0;            			    // MUX 0, AVCC als Ref.
  
    #if (defined(__AVR_ATmega32__))
        /* SFIOR (DB S. 218f. )
         * - free running mode (default) */
        SFIOR  &=  ~((1<<ADTS2)|(1<<ADTS1)|(1<<ADTS0));
         
    #elif (defined(__AVR_ATmega644__))
        // ADC free running (ATmega644,default)
        ADCSRB &=  ~((1<<ADTS2)|(1<<ADTS1) |(1<<ADTS0));
        
    #endif

}


/* ADC deinitialisieren, Interrupt abschalten */
void adc_deinit(void)
{
    adc_stop();
}


/* Start erste Konvertierung */
void adc_start(void)
{
	ADCSRA |= (1<<ADSC);
	sei();
}  


void adc_stop(void)
{
	ADCSRA &=  ~((1<<ADEN) | (1<<ADIE));
}
