﻿/*
 * dcmotor.c
 *
 * Steuerung eines oder mehrerer DC-Motoren mit je einer Doppelbruecke, z.B. LM298, L6203
 *
 * Signale: PinA, PinB, Enable
 *
 * Created: 31.08.2015 12:55:34
 *  Author: Nicolas
 */ 


#include "dcmotor.h"
#include "../../common/src_base/assert.h"
#include <stdlib.h>




// Unvollstaendigen Typ vervollstaendigen
struct dcmotorstruct 
{
    GPIO_t * GPIOA; // Beide Pins der Vollbruecke
    GPIO_t * GPIOB;
    GPIO_t * GPIOEn; 
    IOPin_t PinA;
    IOPin_t PinB;
    IOPin_t PinEn;
};



/* Initialisierung: Pins auf Ausgang und struct mit Motordaten zurueckgeben */
dcmotor_t * dcmotor_init(dcmotorinit_t * pinitdata)
{
    dcmotor_t *pmotor = (dcmotor_t *) malloc(sizeof(dcmotor_t));
    pmotor->PinA   = pinitdata->PinA;
    pmotor->PinB   = pinitdata->PinB;
	pmotor->PinEn  = pinitdata->PinEn;
    pmotor->GPIOA  = pinitdata->GPIOA;
    pmotor->GPIOB  = pinitdata->GPIOB;
	pmotor->GPIOEn = pinitdata->GPIOEn;
        
    // Pins auf Ausgang (das dauert ein wenig)
    io_setOutput(pmotor->GPIOA, pmotor->PinA);
    io_setOutput(pmotor->GPIOB, pmotor->PinB);
    io_setOutput(pmotor->GPIOEn, pmotor->PinEn);        
        
    // Pins high bis auf Enable
    io_setBit(pmotor->GPIOA, pmotor->PinA);
    io_setBit(pmotor->GPIOB, pmotor->PinB);
    io_setBit(pmotor->GPIOEn, pmotor->PinEn);

    return pmotor;
}
    
	
/* Deinitialisierung */	
void dcmotor_deinit(dcmotor_t * pmotor)
{
        
    // Pins auf Eingang und Low (das dauert ein wenig)
    io_setInput(pmotor->GPIOA, pmotor->PinA);
    io_setInput(pmotor->GPIOB, pmotor->PinB);
    io_setInput(pmotor->GPIOEn, pmotor->PinEn);
        
    io_clearBit(pmotor->GPIOA, pmotor->PinA);
    io_clearBit(pmotor->GPIOB, pmotor->PinB);
    io_clearBit(pmotor->GPIOEn, pmotor->PinEn);

    // Motor-Struct loeschen
    free(pmotor);
}        
	
	

/* Motordrehung 
 * pmotor: Motor-Struct
 * state: gewuenschte Drehrichtung
 *
 * Da hier ein Vollbrueckentreiber und keine zwei Halbbruecken angesteuert werden, gibt 
 * es kein Problem damit, in welcher Kombination die beiden Brueckenteile Low- und 
 * High-Pegel aufweisen. */
void dcmotor_control(dcmotor_t * pmotor, motion_t state)
{
	switch(state) 
    {
    	case motor_left: 
			/* In diesem Kontext werden die SetBit/ClearBit-Funktionen nicht
		     * zu einem einzelnen SBI/CBI-Befehl optimiert. Damit kann man hier
			 * gut leben, da diese Funktion alles andere als zeitkritisch ist. */
			io_setBit(pmotor->GPIOEn,pmotor->PinEn);
			io_setBit(pmotor->GPIOA, pmotor->PinA);
			io_clearBit(pmotor->GPIOB, pmotor->PinB);
			break;

        case motor_runout_left:
            /* One-Phase-Chopping des L6203 nach Fig. 8b im DB */            
			io_setBit(pmotor->GPIOEn,pmotor->PinEn);
			io_setBit(pmotor->GPIOA, pmotor->PinA);
			io_setBit(pmotor->GPIOB, pmotor->PinB);
            break;
			        
		case motor_stop:
			io_clearBit(pmotor->GPIOEn,pmotor->PinEn);
			io_clearBit(pmotor->GPIOA, pmotor->PinA);
			io_clearBit(pmotor->GPIOB, pmotor->PinB);
			break;

        case motor_runout_right:
            /* One-Phase-Chopping des L6203 nach Fig. 8b im DB */
            io_setBit(pmotor->GPIOEn,pmotor->PinEn);
            io_setBit(pmotor->GPIOA, pmotor->PinA);
            io_setBit(pmotor->GPIOB, pmotor->PinB);
            break;

		case motor_right: 
			io_setBit(pmotor->GPIOEn,pmotor->PinEn);
			io_clearBit(pmotor->GPIOA, pmotor->PinA);
			io_setBit(pmotor->GPIOB, pmotor->PinB);
			break;
                                
		default: // Fehlerfall
			io_clearBit(pmotor->GPIOEn, pmotor->PinEn);
			io_clearBit(pmotor->GPIOA, pmotor->PinA);
			io_clearBit(pmotor->GPIOB, pmotor->PinB);
			error("Invalid state - system halted");
	}
}
