/*
 * ftm_eeprom.c
 *
 * Aktuelle Positionsdaten/Pulszahlen im EEPROM sichern oder wiederherstellen
 *
 * Created: 23.02.2016 19:41:15
 *  Author: Nicolas
 */ 



#include "rot_variables.h"
#include <stdint.h>
#include <string.h>
#include <stdbool.h>
#include "../../common/src_avrcom/platform.h"
#include "../../common/src_base/datatypes.h"
#include "../../common/src_base/assert.h"
#include "../../common/src_misc/crc.h"
#include "../../project/src_ui/popups_prj.h"

#include "../../common/src_encoder/key.h"


#ifdef AVR
    #include <avr/eeprom.h>

#elif defined(__MINGW64__) || defined(__linux__)
    #include "../../common/src_eeprom/eeprom_24Cx.h"
    typedef uint32_t uint24_t;
    
#else
    #error "unknown MCU"
#endif

// FIXME: Nur zum Testen!
#include "../../ioMapping.h"
#include "../../common/src_avrcom/io.h"
#include "../../common/src_avrcom/delay.h"


// Anfangs- und Endadresse des erlaubten Bereichs
#define MEM_START_BYTE 128U
#define MEM_END_BYTE 1023U



/* EEPROM-Daten (24 Bit Breite) */
__extension__ typedef union
{
    // Darstellung zum Fuellen:
    struct
    {
        unsigned fill  : 4;     // Freier Platz fuer CRC
        unsigned pulse1: 10;    // Puls + Offset
        unsigned pulse0: 10;    // Puls + Offset
    } STRUCT_ATTRIBUTE_MAX_PACKED;
    
    // Darstellung fuer CRC-Berechnung:
    struct
    {
        
        unsigned      crc: 4;   // CRC
        unsigned long dat: 20;  // Nutzdaten
    } STRUCT_ATTRIBUTE_MAX_PACKED;
    
    // Darstellung zum sichern im EEPROM:
    uint24_t blk;
    uint8_t byte[3];
} Eedata_t;

#define MEMSLOT_SIZE sizeof(Eedata_t)
#define N_MEMSLOTS ( (MEM_END_BYTE-MEM_START_BYTE+1)/MEMSLOT_SIZE )
#define MEM_LASTSLOT (MEM_START_BYTE+MEMSLOT_SIZE*(N_MEMSLOTS-1))



#ifdef PLATFORM_AVR
    static_assert(sizeof(Eedata_t)==3,"Unexpected width of struct");
    static_assert(MEMSLOT_SIZE == 3,"Memslot size = 3");
#elif defined(PLATFORM_PC)
    // wegen uint24_t == uint32_t
    static_assert(sizeof(Eedata_t)==4,"Unexpected width of struct");
    
#else
    #error "unknown platform"    
    
#endif







/* 10-Bit-Daten pulse0 und pulse1 in 24-Bit-Variable packen und mit CRC4-Pruefsumme 
 * versehen. */
Eedata_t packEedat(int16_t pulse0, int16_t pulse1)
{
    Eedata_t eedat = {.blk = 0};
        
    // Nutzdaten vorzeichenlos machen    
    pulse0 += 255; 
    pulse1 += 255;
    
    // Winkelbereich: -240 ... 960
    // Pulse pro 360: 1 ... 255
    // Pulsbereich: -170 ... 680
    // Sicherheitsabstand zu 0 ... 1024 wird auf beide Seiten verteilt
    // -> Offset = 255;
    assert(pulse0>=0 && pulse0<1024 );
    assert(pulse1>=0 && pulse1<1024 );    
    
    // EEPROM-Daten in 24Bit breites Bitfeld packen:
    // | 10 Bit | 10 Bit | 4 Bit CRC |
    eedat.pulse0 = pulse0;
    eedat.pulse1 = pulse1;
    
    eedat.crc = crc4_calc32(eedat.dat,5);
    
    #if 0
        // Es ist einfacher, den errechneten CRC zu vergleichen, als den CRC ueber den 
        // ganzen Block zu bilden (da unterschiedliche Breite von uint24_t) ...
        if ( crc4_calc32( eedat.dat, 5 ) != eedat.crc )
        {
            error( "CRC calculation for EEPROM data failed" );
        }

    #else
        // ... aber das geht natuerlich auch.
        if ( crc4_calc32( eedat.blk, 6 ))
        {
            error( "CRC calculation for EEPROM data failed" );
        }

    #endif
    
    return eedat;    
}






/* Ersten nicht-leeren Speicherplatz finden. 
 * Fuer diese Funktion muss man etwas Zeit mitbringen. Sie benoetigt ca. 4,5 
 * Millisekunden, um das EEPROM komplett zu durchsuchen.
 * *block: Gibt gefundenen EEPROM-Block zurueck
 * return: 0: kein Inhalt != 0xFF gefunden
 *         anderer Wert: Startadresse  */
uint16_t find_slot(uint24_t *block)
{
    uint8_t blk[MEMSLOT_SIZE];
    uint16_t slotaddr = 0; // Der Wert 0 ist ungueltig, da MEM_START_BYTE > 0 
                           // markiert damit ein "nicht gefunden"
    
    for (uint16_t addr = MEM_START_BYTE; addr <= MEM_LASTSLOT; addr+=MEMSLOT_SIZE)
    {
        // Auf dem AVR sind die EEPROM-Routinen zeigerbasiert
        eeprom_read_block(blk,(uint8_t*) addr,MEMSLOT_SIZE);
        
        if ( (blk[0] == 0xFF) && (blk[1] == 0xFF) && (blk[2] == 0xFF) )
        {
            // unbelegter Platz enhaelt nur 0xFF
        }
        else
        {
            slotaddr = addr;
            break;
        }
    }    
    
    if (block != NULL)
    {
        memcpy(block,blk,MEMSLOT_SIZE);
    }
    
    return slotaddr;
}





static uint16_t gl_actMemslot;

/* Im EERPOM hinterlegten Pulszaehler wiederherstellen. Das kann wegen der Durchsuchung 
 * des EEPROMs durchaus lange dauern. */
uint_fast8_t restorePulse(void)
{
    uint_fast8_t stat = EXIT_FAILURE;
    Eedata_t eedat;
    bool isValid;
    
    uint16_t addr = find_slot(&eedat.blk);
    
    // Testen nach dem auslesen
    isValid = (crc4_calc32(eedat.blk,6) == 0);
    
    if (addr == 0)
    {
        // EEPROM-Daten konnten nicht gefunden werden -> Defaultwerte annehmen und 
        // Messagebox darstellen
        popup_pulseDataNotFound(0U);
    }
    else if (!isValid)
    {
        // EEPROM-Daten konnten nicht richtig gelesen werden -> Defaultwerte annehmen und 
        // Messagebox darstellen
        popup_pulseDataNotFound(1U);
    }
    else
    {
        // Nutzdaten extrahieren
        Glcstat.mot[0].pulse = eedat.pulse0-255;
        Glcstat.mot[1].pulse = eedat.pulse1-255;
        stat=EXIT_SUCCESS;
    }
    
    gl_actMemslot = addr;
    return stat;
}



/* Diese Funktion darf nicht zu oft aufgerufen werden - sie verschleisst sonst das 
 * EEPROM! */
void savePulse(void)
{
    uint16_t lastmemslot = gl_actMemslot;
    Eedata_t eedat_new, eedat_old;
    
    // Daten zusammenpacken
    eedat_new = packEedat(Glcstat.mot[0].pulse, Glcstat.mot[1].pulse);
    
    // Allte EEPROM-Daten lesen
    eeprom_read_block(&eedat_old.blk,(uint8_t*) lastmemslot,MEMSLOT_SIZE);
    
    if (eedat_new.blk == eedat_old.blk)
    {
        // Wenn die letzten Positionsdaten identisch sind, muss nichts mehr gemacht 
        // werden.
        ;
    }
    else
    {
        if (gl_actMemslot == 0)
        {
            // In Memslot 0 sind keine gueltigen EEPROM-Daten vorhanden ...
            gl_actMemslot = MEM_START_BYTE;
            
            // ...also wird an der ersten gueltigen Stelle geschrieben.
            eeprom_update_block(&eedat_new.blk,(uint8_t*) gl_actMemslot,MEMSLOT_SIZE);
        }
        else
        {
            // Einen Speicherplatz weiterruecken ...
            gl_actMemslot += MEMSLOT_SIZE;
            if (gl_actMemslot+MEMSLOT_SIZE > MEM_END_BYTE)
            {
                gl_actMemslot = MEM_START_BYTE;
            }
            
            // ... an der neuen Stelle schreiben ...
            eeprom_update_block(&eedat_new.blk,(uint8_t*) gl_actMemslot,MEMSLOT_SIZE);
            
            // ...und an der alten Stelle mit loeschen, d.h. mit 0xFFFFFF ueberschreibben
            // (falls das ueberhaupt ein gueltiger Slot war).
            if (lastmemslot)
            {
                uint24_t tmp = (uint24_t) ~0UL;
                eeprom_update_block(&tmp,(uint8_t*) lastmemslot,MEMSLOT_SIZE);
                
            }
            else
            {
                // ... wobei dieser Fall nie auftreten duerfte, da schon oben abgedeckt.
                error("Das darf gar nicht passieren");   
            }                
        }            
    }
}



/* ************************************************************************************ *
 *   Tests                                                                              *
 * ************************************************************************************ */

/* Testen, ob gepackte EEPROM-Daten fehlerfrei wiederhergestellt werden koennen. */ 
void test_eepack(void) 
{
    Eedata_t eedat;
    
    
    for (int16_t i = -170; i<681; i++)
    {
        for (int16_t j = -170; j<681; j++)
        {
            eedat = packEedat(i, j);
            
            int16_t npulse0 = eedat.pulse0-255;
            int16_t npulse1 = eedat.pulse1-255;
            
            if (npulse0 != i)
            {
                error("pulse 0");
            }
            if (npulse1 != j)
            {
                error("pulse 1");
            }
        }
    }
    printf("test passed");
}


/* Kleine Testfunktion, um Rohdaten (wie beim Systemstart) zu lesen - "restorePulse" zum 
 * mitlesen */
void test_ftm_eep_read(void)
{
    printf_P(PSTR("\fTest EE-Read"));
 
    uint_fast8_t stat = EXIT_FAILURE;
    Eedata_t eedat;
    bool isValid;
 
    uint16_t addr = find_slot(&eedat.blk);
    printf_P(PSTR("\nAddr=%u\a"),addr);
    
    // Testen nach dem auslesen
    isValid = (crc4_calc32(eedat.blk,6) == 0);
    printf_P(PSTR("valid=%u\a"),isValid); 
     
    
    printf_P(PSTR("\npul0=%i, pul1=%i,crc=%x\a"),\
        eedat.pulse0-255,eedat.pulse1-255,eedat.crc);  
     
    while (!key_getstroke());
    
    return;
 
     if (addr == 0)
     {
         // EEPROM-Daten konnten nicht gefunden werden -> Defaultwerte annehmen und 
         // Messagebox darstellen
         popup_pulseDataNotFound(0U);
     }
     else if (!isValid)
     {
         // EEPROM-Daten konnten nicht richtig gelesen werden -> Defaultwerte annehmen und 
         // Messagebox darstellen
         popup_pulseDataNotFound(1U);
     }
     else
     {
         Glcstat.mot[0].pulse = eedat.pulse0-255;
         Glcstat.mot[1].pulse = eedat.pulse1-255;
         stat=EXIT_SUCCESS;
     }
 
     gl_actMemslot = addr;
    
    
}