/*
 * pulsecounter.c
 *
 * Liest Taktflanken vom Antennenrotor ein und zaehlt sie
 *
 * Schreibt in Glcstat
 *
 * Created: 30.01.2016 20:02:08
 *  Author: Nicolas
 */ 


#include "pulsecounter.h"
#include "../../ioMapping.h"
#include "rot_variables.h"




/* Struct fr debounce initialisieren */
void debounce_init(debounce_t * Dbdat)
{
    Dbdat->inpstat = 0x55; // Garantiert maximal unentprellter Zustand
    Dbdat->oldstat = 0x2;  // Garantiert ungueltiger Pin-Status
}


/* Pin Entprellen: Steigende und fallende Flanken feststellen 
 * Beliebig viele Pins koennen entprellt werden.
 * 
 * Aufruf direkt an Portpin z.B.: debounce(dbdat, io_bitIsSet(pGPIOx,Pin) );
 *
 * Solange noch kein stabiler, entprellter Zustand hergestellt ist, sind die
 * Rueckgabewerte Zwangslaeufig ungueltig. Das ist mindestens waehrend der 
 * ersten 7 Funktionsaufrufe der Fall.
 *
 * Dbdat: Interne Daten zum Entprellen
 * dbval: Zu entprellende Pin-Werte
 * return: 0: Low-Pegel, 1: High-Pegel, 2: Fallende Flanke, 3: Steigende Flanke  */
static inline __attribute__((always_inline))
 uint_fast8_t debounce(debounce_t * Dbdat, uint_fast8_t dbval)
{
    
    uint_fast8_t output;      // Flanke, High, Low
    uint_fast8_t oldstat, newstat;
    
    oldstat = Dbdat->oldstat;
    newstat = oldstat;

    // Zustand Entprellen
    Dbdat->inpstat = (Dbdat->inpstat<<1) | (dbval!=0);    

    // Fester Zustand ist dann, wenn 8mal abgefragt das gleiche Ergebnis kommt. 
    // Erst dann wird newstat geaendert.
    if (Dbdat->inpstat == 0x00) 
    {
        newstat = 0;
    }
    else if (Dbdat->inpstat == 0xFF) 
    {
        newstat = 1;
    }
    
    
    // Flanken detektieren
    switch(newstat) 
    {
        case 0:
             // schon vorher low oder fallende Flanke ?
             output = oldstat==1 ? falling_edge : low;
             break;
        
        case 1:
            // schon vorher high oder steigende Flanke ?
            output = oldstat     ? high : raising_edge;
            break;      
        
        default: // Ab 2: ungueltige Zustaende
            output = invalid;
    }
 
    
    Dbdat->oldstat = newstat;
    
    return output;
}


// Debounce-Daten fuer die beiden Input-Pins
static debounce_t Dbdat0;
static debounce_t Dbdat1;



/* Initialisierung muss vor dem ADC-Takt erfolgen */
void pulsecounter_init(void)
{
    debounce_init(&Dbdat0);
    debounce_init(&Dbdat1);
}


/* Pulse zaehlen - gehoert mindestens alle 10ms aufgerufen 
 *
 * Geaendert werden die Pulszaehler Glcstat.mot[0].pulse und Glcstat.mot[1].pulse */

/* Eigentlich schreit diese Funktion nach "static inline" - die ISR sind allerdings auch
 * so schon ausreichend schnell, so dass keine Optimierung in dieser Richtung erforderlich
 * ist. */
void count_pulses(void)
{
    uint_fast8_t db[2];
    
    db[0] = debounce(&Dbdat0,io_bitIsSet(PULSE_IN1_GPIO,PULSE_IN1_Pin));
    db[1] = debounce(&Dbdat1,io_bitIsSet(PULSE_IN2_GPIO,PULSE_IN2_Pin));    
    
    for (uint8_t i = 0; i<2; i++)
    {
    
        if (db[i] == falling_edge) 
        {
            // Die Motor-Drehrichtung laesst sich am einfachsten am tatsaechlichen
            // Pin-Output feststellen
            if (Glcstat.mot[i].motion < 0 ) 
            {
                Glcstat.mot[i].pulse--;    
            }
            else if (Glcstat.mot[i].motion > 0 ) 
            {
                Glcstat.mot[i].pulse++;    
            }
            else 
            {
                // Diesen Puls ignorieren - er stammt vom Abschalten des Motors
            }
        }
    }
}

