/*
 * rot_settings.h
 *
 * Harte und EEPROM-Einstellungen fuer die Rotorsteuerung
 *
 * 
 * Created: 27.08.2015
 *  Author: Nicolas
 */ 



#ifndef ROT_SETTINGS_H
    #define ROT_SETTINGS_H

	#define MAGIC16 0xA55Au


    /* Harte Einstellungen */
    #define MAX_CURRENT_MA 4000UL  // mA
    
    
    /* Umrechnung ADC-Werte in Strom [mA] 
     *  I [mA] = (adcval * DIGIT_CURRENT_ENUM)/DIGIT_CURRENT_DENOM  
     * Werte sind ausgemessen in Zielschaltung */
    #define DIGIT_CURRENT_ENUM 12U   // Zaehler
    #define DIGIT_CURRENT_DENOM 1U   // Nenner, muss Zweierpotenz sein!

    #define MAX_CURRENT_DIGIT ((MAX_CURRENT_MA * DIGIT_CURRENT_DENOM)/DIGIT_CURRENT_ENUM)

    // Nach dieser Zeit [ms] ohne Puls und Stromaufnahme wird der Rotor als nicht 
    // vorhanden markiert
    #define NO_ROTOR_TIMEOUT_MS 5000U   
    
    // Nach dieser Zeit [ms] des Rotorstillstandes wird die aktuelle Position im EEPROM 
    // gesichert
    #define SAVE_EEPROM_AFTER_MS 3000U  

    #include <stdint.h>


    // Daten je Motor/Rotor
    typedef struct
    {
        int16_t limitDeg;          // Himmelsrichtung des virtuellen Anschlags  []       
        int16_t stepoverDeg;       // Erlaubter Ueberschneidungsbereich []    
        int16_t pulsePerRev;       // Anzahl Impulse pro Rotordrehung
        int8_t  dir;               // Logik der Drehrichtung (-1, 1)
    } motorset_t;


    // Konfigurationsdaten
    typedef struct 
    {
        uint16_t magic;            // Magic Number zur Identifikation des Anfangs und 
                                   // damit der CRC bei leeren Settings nicht Null wird
                                   
        uint8_t screen;            // Darstellungsmodus:
                                   //    0: grafische Anzeige, rechts/links/stop
                                   //    1: grafische Anzeige, Winkel
                                   //    2: Textanzeige, rechts/links/stop
                                   //    3: Textanzeige, Winkel     
        
        motorset_t mot[2];         // Daten fuer die einzelnen Motoren
                                   
        uint16_t iMax;             // Darueber Notabschaltung wegen Blockade
        uint16_t runoutTimeMs;     // Zeit fuer den Motorauslauf vor 
                                   // Drehrichtungsumkehr [Millisekunden]
    } rot_settings_t;

    extern rot_settings_t GlSetting;

  

    /* ******************************************************************************** */
    /*  Max, min und Defaultwerte                                                       */
    /* ******************************************************************************** */
    /* Max- und Min-Werte limitieren die Einstellmglichkeiten im Einstellmen.
     * Defaultwerte werden genutzt falls 
     *   a) aus dem EEPROM ungueltige Daten gelesen werden oder
     *   b) das Geraet zurueckgesetzt wird oder
     *   c) die Firmware frisch geflash wurde.
     */
    #define GLOBAL_SETTINGS_DEFAULT \
    {\
        .magic = MAGIC16, \
        .screen = 1,\
        .mot[0] = {.limitDeg=0, .stepoverDeg = 45, .pulsePerRev = 158U, .dir = 1,},\
        .mot[1] = {.limitDeg=0, .stepoverDeg = 45, .pulsePerRev = 158U, .dir = 1,},\
        .iMax = 4000U,\
        .runoutTimeMs = 1000U\
    }
  
  
    // Funktionen umbenennen fuer Zusammenarbeit mit Standard-EEPROM-Routinen
    #define rot_settings_read     settings_read
    #define rot_settings_write    settings_write
    #define rot_settings_default  settings_default      
    
  
    // Funktionen
    uint_fast8_t rot_settings_init(void);
    uint_fast8_t rot_settings_read(void);
    uint_fast8_t rot_settings_write(void);
    uint_fast8_t rot_settings_default(void);

#endif 
